"""opentrons.protocol_api.labware: classes and functions for labware handling

This module provides things like :py:class:`Labware`, and :py:class:`Well`
to encapsulate labware instances used in protocols
and their wells. It also provides helper functions to load and save labware
and labware calibration offsets. It contains all the code necessary to
transform from labware symbolic points (such as "well a1 of an opentrons
tiprack") to points in deck coordinates.
"""

from __future__ import annotations

import logging

from itertools import dropwhile
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Dict,
    Optional,
    Tuple,
    cast,
    Sequence,
    Mapping,
    Union,
    Literal,
)

from opentrons_shared_data.labware.types import (
    LabwareDefinition,
    LabwareDefinition2,
    LabwareParameters2,
    LabwareParameters3,
)

from opentrons.types import (
    Location,
    Point,
    NozzleMapInterface,
    MeniscusTrackingTarget,
    Mount,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import (
    requires_version,
    APIVersionError,
    UnsupportedAPIError,
)
from opentrons.protocol_engine.types import LiquidTrackingType

# TODO(mc, 2022-09-02): re-exports provided for backwards compatibility
# remove when their usage is no longer needed
from opentrons.protocols.labware import (  # noqa: F401
    get_labware_definition as get_labware_definition,
    verify_definition as verify_definition,
    save_definition as save_definition,
)

from . import validation
from ._liquid import Liquid
from ._types import OffDeckType
from .core import well_grid
from .core.engine import (
    ENGINE_CORE_API_VERSION,
    SET_OFFSET_RESTORED_API_VERSION,
)
from .core.labware import AbstractLabware
from .core.module import AbstractModuleCore
from .core.core_map import LoadedCoreMap
from .core.legacy.legacy_labware_core import LegacyLabwareCore
from .core.legacy.legacy_well_core import LegacyWellCore
from .core.legacy.well_geometry import WellGeometry


if TYPE_CHECKING:
    from .core.common import LabwareCore, WellCore, ProtocolCore
    from .protocol_context import ModuleTypes


_log = logging.getLogger(__name__)


_IGNORE_API_VERSION_BREAKPOINT = APIVersion(2, 13)
"""API version after which to respect... the API version setting.

At this API version and below, `Labware` objects were always
erroneously constructed set to MAX_SUPPORTED_VERSION.
"""


class TipSelectionError(Exception):
    pass


class OutOfTipsError(Exception):
    pass


class Well:
    """
    The Well class represents a single well in a :py:class:`Labware`. It provides parameters and functions for three major uses:

        - Calculating positions relative to the well. See :ref:`position-relative-labware` for details.

        - Returning well measurements. See :ref:`new-labware-well-properties` for details.

        - Specifying what liquid should be in the well at the beginning of a protocol. See :ref:`labeling-liquids` for details.
    """

    def __init__(self, parent: Labware, core: WellCore, api_version: APIVersion):
        if api_version <= _IGNORE_API_VERSION_BREAKPOINT:
            api_version = _IGNORE_API_VERSION_BREAKPOINT

        self._parent = parent
        self._core = core
        self._api_version = api_version

    @property
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property
    @requires_version(2, 0)
    def parent(self) -> Labware:
        """The :py:class:`.Labware` object that the well is a part of."""
        return self._parent

    @property
    @requires_version(2, 0)
    def has_tip(self) -> bool:
        """Whether this well contains an unused tip.

        From API v2.2 on:

        - Returns ``False`` if:

          - the well has no tip present, or
          - the well has a tip that's been used by the protocol previously

        - Returns ``True`` if the well has an unused tip.

        Before API v2.2:

        - Returns ``True`` as long as the well has a tip, even if it is used.

        Always ``False`` if the parent labware isn't a tip rack.
        """
        return self._core.has_tip()

    @has_tip.setter
    def has_tip(self, value: bool) -> None:
        _log.warning(
            "Setting the `Well.has_tip` property manually has been deprecated"
            " and will raise an error in a future version of the Python Protocol API."
        )

        self._core.set_has_tip(value)

    @property
    def max_volume(self) -> float:
        """The maximum volume, in µL, that the well can hold.

        This amount is set by the JSON labware definition, specifically the ``totalLiquidVolume`` property of the particular well.
        """
        return self._core.get_max_volume()

    @property
    def geometry(self) -> WellGeometry:
        if isinstance(self._core, LegacyWellCore):
            return self._core.geometry
        raise UnsupportedAPIError(api_element="Well.geometry")

    @property
    @requires_version(2, 0)
    def diameter(self) -> Optional[float]:
        """
        The diameter, in mm, of a circular well. Returns ``None``
        if the well is not circular.
        """
        return self._core.diameter

    @property
    @requires_version(2, 9)
    def length(self) -> Optional[float]:
        """
        The length, in mm, of a rectangular well along the x-axis (left to right).
        Returns ``None`` if the well is not rectangular.
        """
        return self._core.length

    @property
    @requires_version(2, 9)
    def width(self) -> Optional[float]:
        """
        The width, in mm, of a rectangular well along the y-axis (front to back).
        Returns ``None`` if the well is not rectangular.
        """
        return self._core.width

    @property
    @requires_version(2, 9)
    def depth(self) -> float:
        """
        The depth, in mm, of a well along the z-axis, from the very top of the well to
        the very bottom.
        """
        return self._core.depth

    @property
    def display_name(self) -> str:
        """A human-readable name for the well, including labware and deck location.

        For example, "A1 of Corning 96 Well Plate 360 µL Flat on slot D1". Run log
        entries use this format for identifying wells. See
        :py:meth:`.ProtocolContext.commands`.
        """
        return self._core.get_display_name()

    @property
    @requires_version(2, 7)
    def well_name(self) -> str:
        """A string representing the well's coordinates.

        For example, "A1" or "H12".

        The format of strings that this property returns is the same format as the key
        for :ref:`accessing wells in a dictionary <well-dictionary-access>`.
        """
        return self._core.get_name()

    @requires_version(2, 0)
    def top(self, z: float = 0.0) -> Location:
        """
        :param z: An offset on the z-axis, in mm. Positive offsets are higher and
            negative offsets are lower.

        :return: A :py:class:`~opentrons.types.Location` corresponding to the
            absolute position of the top-center of the well, plus the ``z`` offset
            (if specified).
        """
        return Location(self._core.get_top(z_offset=z), self)

    @requires_version(2, 0)
    def bottom(self, z: float = 0.0) -> Location:
        """
        :param z: An offset on the z-axis, in mm. Positive offsets are higher and
            negative offsets are lower.

        :return: A :py:class:`~opentrons.types.Location` corresponding to the
            absolute position of the bottom-center of the well, plus the ``z`` offset
            (if specified).
        """
        return Location(self._core.get_bottom(z_offset=z), self)

    @requires_version(2, 0)
    def center(self) -> Location:
        """
        :return: A :py:class:`~opentrons.types.Location` corresponding to the
            absolute position of the center of the well (in all three dimensions).
        """
        return Location(self._core.get_center(), self)

    @requires_version(2, 21)
    def meniscus(
        self, z: float = 0.0, target: Literal["start", "end", "dynamic"] = "end"
    ) -> Location:
        """
        :param z: An offset on the z-axis, in mm. Positive offsets are higher and
            negative offsets are lower.
        :param target: The relative position of the liquid meniscus inside the well to target when performing a liquid handling operation.

        :return: A :py:class:`~opentrons.types.Location` corresponding to the liquid meniscus, plus a target position and ``z`` offset as specified.

        """
        return Location(
            point=Point(x=0, y=0, z=z),
            labware=self,
            _meniscus_tracking=MeniscusTrackingTarget(target),
        )

    @requires_version(2, 8)
    def from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """
        Specifies a :py:class:`~opentrons.types.Point` based on fractions of the
        distance from the center of the well to the edge along each axis.

        For example, ``from_center_cartesian(0, 0, 0.5)`` specifies a point at the
        well's center on the x- and y-axis, and half of the distance from the center of
        the well to its top along the z-axis. To move the pipette to that location,
        construct a :py:class:`~opentrons.types.Location` relative to the same well::

            location = types.Location(
                plate["A1"].from_center_cartesian(0, 0, 0.5), plate["A1"]
            )
            pipette.move_to(location)

        See :ref:`points-locations` for more information.

        :param x: The fraction of the distance from the well's center to its edge
            along the x-axis. Negative values are to the left, and positive values
            are to the right.
        :param y: The fraction of the distance from the well's center to its edge
            along the y-axis. Negative values are to the front, and positive values
            are to the back.
        :param z: The fraction of the distance from the well's center to its edge
            along the x-axis. Negative values are down, and positive values are up.

        :return: A :py:class:`~opentrons.types.Point` representing the specified
            position in absolute deck coordinates.

        .. note:: Even if the absolute values of ``x``, ``y``, and ``z`` are all less
            than 1, a location constructed from the well and the result of
            ``from_center_cartesian`` may be outside of the physical well. For example,
            ``from_center_cartesian(0.9, 0.9, 0)`` would be outside of a cylindrical
            well, but inside a square well.

        """
        return self._core.from_center_cartesian(x, y, z)

    @requires_version(2, 14)
    def load_liquid(self, liquid: Liquid, volume: float) -> None:
        """
        Load a liquid into a well.

        :param Liquid liquid: The liquid to load into the well.
        :param float volume: The volume of liquid to load, in µL.

        .. deprecated:: 2.22
            Use :py:meth:`.Labware.load_liquid`, :py:meth:`.Labware.load_liquid_by_well`, or :py:meth:`.Labware.load_empty` instead.

        """
        self._core.load_liquid(
            liquid=liquid,
            volume=volume,
        )

    @requires_version(2, 21)
    def current_liquid_height(self) -> LiquidTrackingType:
        """Get the current liquid height in a well."""
        return self._core.current_liquid_height()

    @requires_version(2, 21)
    def current_liquid_volume(self) -> LiquidTrackingType:
        """Get the current liquid volume in a well."""
        return self._core.get_liquid_volume()

    @requires_version(2, 24)
    def volume_from_height(self, height: LiquidTrackingType) -> LiquidTrackingType:
        """Return the volume contained in a well at any height."""
        return self._core.volume_from_height(height)

    @requires_version(2, 24)
    def height_from_volume(self, volume: LiquidTrackingType) -> LiquidTrackingType:
        """Return the height in a well corresponding to a given volume."""
        return self._core.height_from_volume(volume)

    @requires_version(2, 21)
    def estimate_liquid_height_after_pipetting(
        self,
        mount: Mount | str,
        operation_volume: float,
    ) -> LiquidTrackingType:
        """Check the height of the liquid within a well.

        :returns: The height, in mm, of the liquid from the deck.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """

        projected_final_height = self._core.estimate_liquid_height_after_pipetting(
            operation_volume=operation_volume, mount=mount
        )
        return projected_final_height

    def _from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """
        Private version of from_center_cartesian. Present only for backward
        compatibility.
        """
        _log.warning(
            "This method is deprecated. Please use 'from_center_cartesian' instead."
        )
        return self.from_center_cartesian(x, y, z)

    def __repr__(self) -> str:
        return self._core.get_display_name()

    def __eq__(self, other: object) -> bool:
        """
        Assuming that equality of wells in this system is having the same
        absolute coordinates for the top.
        """
        if not isinstance(other, Well):
            return NotImplemented
        return self.top().point == other.top().point

    def __hash__(self) -> int:
        return hash(self.top().point)


class Labware:
    """
    This class represents a piece of labware.

    Labware available in the API generally fall under two categories.

      - Consumable labware: well plates, tubes in racks, reservoirs, tip racks, etc.
      - Adapters: durable items that hold other labware, either on modules or directly
        on the deck.

    The ``Labware`` class defines the physical geometry of the labware
    and provides methods for :ref:`accessing wells <new-well-access>` within the labware.

    Create ``Labware`` objects by calling the appropriate ``load_labware()`` method,
    depending on where you are loading the labware. For example, to load labware on a
    Thermocycler Module, use :py:meth:`.ThermocyclerContext.load_labware`. To load
    labware directly on the deck, use :py:meth:`.ProtocolContext.load_labware`. See
    :ref:`loading-labware`.

    """

    def __init__(
        self,
        core: AbstractLabware[Any],
        api_version: APIVersion,
        protocol_core: ProtocolCore,
        core_map: LoadedCoreMap,
    ) -> None:
        """
        :param core: The class that implements the public interface
                               of the class.
        :param APIVersion api_level: the API version to set for the instance.
                                     The :py:class:`.Labware` will
                                     conform to this level. If not specified,
                                     defaults to
                                     :py:attr:`.MAX_SUPPORTED_VERSION`.
        """
        if api_version <= _IGNORE_API_VERSION_BREAKPOINT:
            api_version = _IGNORE_API_VERSION_BREAKPOINT

        self._api_version = api_version
        self._core: LabwareCore = core
        self._protocol_core = protocol_core
        self._core_map = core_map

        well_columns = core.get_well_columns()
        self._well_grid = well_grid.create(columns=well_columns)
        self._wells_by_name = {
            well_name: Well(
                parent=self, core=core.get_well_core(well_name), api_version=api_version
            )
            for column in well_columns
            for well_name in column
        }

    @property
    def separate_calibration(self) -> bool:
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.separate_calibration",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
            )

        _log.warning(
            "Labware.separate_calibrations is a deprecated internal property."
            " It no longer has meaning, but will always return `False`"
        )
        return False

    @property
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        """See :py:obj:`.ProtocolContext.api_version`."""
        return self._api_version

    def __getitem__(self, key: str) -> Well:
        return self.wells_by_name()[key]

    @property
    @requires_version(2, 0)
    def uri(self) -> str:
        """A string fully identifying the labware.

        The URI has three parts and follows the pattern ``"namespace/load_name/version"``.
        For example, ``opentrons/corning_96_wellplate_360ul_flat/2``.
        """
        return self._core.get_uri()

    @property
    @requires_version(2, 0)
    def parent(self) -> Union[str, Labware, ModuleTypes, OffDeckType]:
        """Where the labware is loaded.

        This corresponds to the physical object that the labware *directly* rests upon.

        Returns:
            If the labware is directly on the robot's deck, the ``str`` name of the deck slot,
            like ``"D1"`` (Flex) or ``"1"`` (OT-2). See :ref:`deck-slots`.

            If the labware is on a module, a module context.

            If the labware is on a labware or adapter, a :py:class:`Labware`.

            If the labware is off-deck, :py:obj:`OFF_DECK`.

        .. versionchanged:: 2.14
            Return type for module parent changed.
            Formerly, the API returned an internal geometry interface.
        .. versionchanged:: 2.15
            Returns a :py:class:`Labware` if the labware is loaded onto a labware/adapter.
            Returns :py:obj:`OFF_DECK` if the labware is off-deck.
            Formerly, if the labware was removed by using ``del`` on :py:obj:`.deck`,
            this would return where it was before its removal.
        """
        if isinstance(self._core, LegacyLabwareCore):
            # Type ignoring to preserve backwards compatibility
            return self._core.get_geometry().parent.labware.object  # type: ignore

        assert self._protocol_core and self._core_map, "Labware initialized incorrectly"

        labware_location = self._protocol_core.get_labware_location(self._core)

        if isinstance(labware_location, (AbstractLabware, AbstractModuleCore)):
            return self._core_map.get(labware_location)

        return labware_location

    @property
    @requires_version(2, 0)
    def name(self) -> str:
        """The display name of the labware.

        If you specified a value for ``label`` when loading the labware, ``name`` is
        that value.

        Otherwise, it is the :py:obj:`~.Labware.load_name` of the labware.
        """
        return self._core.get_name()

    @name.setter
    def name(self, new_name: str) -> None:
        """Set the labware name.

        .. deprecated: 2.14
            Set the name of labware in `load_labware` instead.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.name setter",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
            )

        # TODO(mc, 2023-02-06): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyLabwareCore)
        cast(LegacyLabwareCore, self._core).set_name(new_name)

    @property
    @requires_version(2, 0)
    def load_name(self) -> str:
        """The API load name of the labware definition."""
        return self._core.load_name

    @property
    @requires_version(2, 0)
    def parameters(self) -> "LabwareParameters2 | LabwareParameters3":
        """Internal properties of a labware including type and quirks."""
        return self._core.get_parameters()

    @property
    @requires_version(2, 0)
    def quirks(self) -> List[str]:
        """Quirks specific to this labware."""
        return self._core.get_quirks()

    # TODO(mm, 2023-02-08):
    # Specify units and origin after we resolve RSS-110.
    # Remove warning once we resolve RSS-109 more broadly.
    @property
    @requires_version(2, 0)
    def magdeck_engage_height(self) -> Optional[float]:
        """Return the default magnet engage height that
        :py:meth:`.MagneticModuleContext.engage` will use for this labware.

        .. warning::
            This currently returns confusing and unpredictable results that do not
            necessarily match what :py:meth:`.MagneticModuleContext.engage` will
            actually choose for its default height.

            The confusion is related to how this height's units and origin point are
            defined, and differences between Magnetic Module generations.

            For now, we recommend you avoid accessing this property directly.
        """
        # Return the raw value straight from the labware definition. For several
        # reasons (see RSS-109), this may not match the actual default height chosen
        # by MagneticModuleContext.engage().
        p = self._core.get_parameters()
        if not p["isMagneticModuleCompatible"]:
            return None
        else:
            return p["magneticModuleEngageHeight"]

    @property
    @requires_version(2, 15)
    def child(self) -> Optional[Labware]:
        """The labware (if any) present on this labware."""
        labware_core = self._protocol_core.get_labware_on_labware(self._core)
        return self._core_map.get(labware_core)

    @requires_version(2, 15)
    def load_labware(
        self,
        name: str,
        label: Optional[str] = None,
        lid: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        """Load a compatible labware onto the labware using its load parameters.

        The parameters of this function behave like those of
        :py:obj:`ProtocolContext.load_labware` (which loads labware directly
        onto the deck). Note that the parameter ``name`` here corresponds to
        ``load_name`` on the ``ProtocolContext`` function.

        :returns: The initialized and loaded labware object.
        """
        labware_core = self._protocol_core.load_labware(
            load_name=name,
            label=label,
            namespace=namespace,
            version=version,
            location=self._core,
        )

        labware = Labware(
            core=labware_core,
            api_version=self._api_version,
            protocol_core=self._protocol_core,
            core_map=self._core_map,
        )

        self._core_map.add(labware_core, labware)

        if lid is not None:
            if self._api_version < validation.LID_STACK_VERSION_GATE:
                raise APIVersionError(
                    api_element="Loading a Lid on a Labware",
                    until_version="2.23",
                    current_version=f"{self._api_version}",
                )
            self._protocol_core.load_lid(
                load_name=lid,
                location=labware_core,
                namespace=namespace,
                version=version,
            )

        return labware

    @requires_version(2, 15)
    def load_labware_from_definition(
        self, definition: LabwareDefinition, label: Optional[str] = None
    ) -> Labware:
        """Load a compatible labware onto the labware using an inline definition.

        :param definition: The labware definition.
        :param str label: An optional special name to give the labware. If specified,
            this is how the labware will appear in the run log, Labware Position
            Check, and elsewhere in the Opentrons App and on the touchscreen.

        :returns: The initialized and loaded labware object.
        """
        load_params = self._protocol_core.add_labware_definition(definition)

        return self.load_labware(
            name=load_params.load_name,
            namespace=load_params.namespace,
            version=load_params.version,
            label=label,
        )

    @requires_version(2, 23)
    def load_lid_stack(
        self,
        load_name: str,
        quantity: int,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        """
        Load a stack of Opentrons Tough Auto-Sealing Lids onto a valid deck location or adapter.

        :param str load_name: A string to use for looking up a lid definition.
            You can find the ``load_name`` for any standard lid on the Opentrons
            `Labware Library <https://labware.opentrons.com>`_.
        :param int quantity: The quantity of lids to be loaded in the stack.
        :param str namespace: The namespace that the lid labware definition belongs to.
            If unspecified, the API will automatically search two namespaces:

              - ``"opentrons"``, to load standard Opentrons labware definitions.
              - ``"custom_beta"``, to load custom labware definitions created with the
                `Custom Labware Creator <https://labware.opentrons.com/create>`__.

            You might need to specify an explicit ``namespace`` if you have a custom
            definition whose ``load_name`` is the same as an Opentrons-verified
            definition, and you want to explicitly choose one or the other.

        :param version: The version of the labware definition. You should normally
            leave this unspecified to let ``load_lid_stack()`` choose a version
            automatically.

        :return:  The initialized and loaded labware object representing the lid stack.
        """
        if self._api_version < validation.LID_STACK_VERSION_GATE:
            raise APIVersionError(
                api_element="Loading a Lid Stack",
                until_version="2.23",
                current_version=f"{self._api_version}",
            )

        load_location = self._core

        load_name = validation.ensure_lowercase_name(load_name)

        result = self._protocol_core.load_lid_stack(
            load_name=load_name,
            location=load_location,
            quantity=quantity,
            namespace=namespace,
            version=version,
        )

        labware = Labware(
            core=result,
            api_version=self._api_version,
            protocol_core=self._protocol_core,
            core_map=self._core_map,
        )
        return labware

    def set_calibration(self, delta: Point) -> None:
        """
        An internal, deprecated method used for updating the labware offset.

        .. deprecated:: 2.14
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.set_calibration()",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
                extra_message="Try using the Opentrons App's Labware Position Check.",
            )
        self._core.set_calibration(delta)

    @requires_version(2, 12)
    def set_offset(self, x: float, y: float, z: float) -> None:
        """Set the labware's position offset.

        The offset is an x, y, z vector in deck coordinates
        (see :ref:`protocol-api-deck-coords`).

        How the motion system applies the offset depends on the API level of the protocol.

        .. list-table::
            :header-rows: 1
            :widths: 1 5

            * - API level
              - Offset behavior
            * - 2.12–2.13
              - Offsets only apply to the exact :py:class:`.Labware` instance.

                If your protocol has multiple instances of the same type of labware,
                you must either use ``set_offset()`` on all of them or none of them.
            * - 2.14–2.17
              - ``set_offset()`` is not available, and the API raises an error.
            * - 2.18--2.22
              -
                - Offsets apply to any labware of the same type, in the same on-deck location.
                - Offsets can't be set on labware that is currently off-deck.
                - Offsets do not follow a labware instance when using :py:meth:`.move_labware`.
            * - 2.23 and newer
              -
                On Flex, offsets can apply to all labware of the same type, regardless of their on-deck location.

        .. note::

            Setting offsets with this method will override any labware offsets set
            by running Labware Position Check in the Opentrons App.

            This method is designed for use with mechanisms like
            :obj:`opentrons.execute.get_protocol_api`, which lack an interactive way
            to adjust labware offsets. (See :ref:`advanced-control`.)

        .. versionchanged:: 2.14
            Temporarily removed.

        .. versionchanged:: 2.18
            Restored, and now applies to labware type–location pairs.
        """
        if (
            self._api_version >= ENGINE_CORE_API_VERSION
            and self._api_version < SET_OFFSET_RESTORED_API_VERSION
        ):
            raise APIVersionError(
                api_element="Labware.set_offset()",
                until_version=f"{SET_OFFSET_RESTORED_API_VERSION}",
                current_version=f"{self._api_version}",
                extra_message="This feature not available in versions 2.14 thorugh 2.17. You can also use the Opentrons App's Labware Position Check.",
            )
        else:
            self._core.set_calibration(Point(x=x, y=y, z=z))

    @property
    @requires_version(2, 0)
    def calibrated_offset(self) -> Point:
        """The front-left-bottom corner of the labware, including its labware offset.

        When running a protocol in the Opentrons App or on the touchscreen, Labware
        Position Check sets the labware offset.
        """
        return self._core.get_calibrated_offset()

    @requires_version(2, 0)
    def well(self, idx: Union[int, str]) -> Well:
        """Deprecated. Use result of :py:meth:`wells` or :py:meth:`wells_by_name`."""
        if isinstance(idx, int):
            return self.wells()[idx]
        elif isinstance(idx, str):
            return self.wells_by_name()[idx]
        else:
            raise TypeError(
                f"`Labware.well` must be called with an `int` or `str`, but got {idx}"
            )

    @requires_version(2, 0)
    def wells(self, *args: Union[str, int]) -> List[Well]:
        """
        Accessor function to navigate a labware top to bottom, left to right.

        i.e., this method returns a list ordered A1, B1, C1…A2, B2, C2….

        Use indexing to access individual wells contained in the list.
        For example, access well A1 with ``labware.wells()[0]``.

        .. note::
            Using args with this method is deprecated. Use indexing instead.

            If your code uses args, they can be either strings or integers, but not a
            mix of the two. For example, ``.wells(1, 4)`` or ``.wells("1", "4")`` is
            valid, but ``.wells("1", 4)`` is not.

        :return: Ordered list of all wells in a labware.
        """
        if not args:
            return list(self._wells_by_name.values())

        elif validation.is_all_integers(args):
            wells = self.wells()
            return [wells[idx] for idx in args]

        elif validation.is_all_strings(args):
            wells_by_name = self.wells_by_name()
            return [wells_by_name[idx] for idx in args]

        else:
            raise TypeError(
                "`Labware.wells` must be called with all `int`'s or all `str`'s,"
                f" but was called with {args}"
            )

    @requires_version(2, 0)
    def wells_by_name(self) -> Dict[str, Well]:
        """
        Accessor function used to navigate through a labware by well name.

        Use indexing to access individual wells contained in the dictionary.
        For example, access well A1 with ``labware.wells_by_name()["A1"]``.

        :return: Dictionary of :py:class:`.Well` objects keyed by well name.
        """
        return dict(self._wells_by_name)

    @requires_version(2, 0)
    def wells_by_index(self) -> Dict[str, Well]:
        """
        .. deprecated:: 2.0
            Use :py:meth:`wells_by_name` or dict access instead.
        """
        _log.warning(
            "wells_by_index is deprecated. Use wells_by_name or dict access instead."
        )
        return self.wells_by_name()

    @requires_version(2, 0)
    def rows(self, *args: Union[int, str]) -> List[List[Well]]:
        """
        Accessor function to navigate through a labware by row.

        Use indexing to access individual rows or wells contained in the nested list.
        On a standard 96-well plate, this will output a list of :py:class:`.Well`
        objects containing A1 through A12.

        .. note::
            Using args with this method is deprecated. Use indexing instead.

            If your code uses args, they can be either strings or integers, but not a
            mix of the two. For example, ``.rows(1, 4)`` or ``.rows("1", "4")`` is
            valid, but ``.rows("1", 4)`` is not.

        :return: A list of row lists.
        """
        if not args:
            return [
                [self._wells_by_name[well_name] for well_name in row]
                for row in self._well_grid.rows_by_name.values()
            ]

        elif validation.is_all_integers(args):
            rows = self.rows()
            return [rows[idx] for idx in args]

        elif validation.is_all_strings(args):
            rows_by_name = self.rows_by_name()
            return [rows_by_name[idx] for idx in args]

        else:
            raise TypeError(
                "`Labware.rows` must be called with all `int`'s or all `str`'s,"
                f" but was called with {args}"
            )

    @requires_version(2, 0)
    def rows_by_name(self) -> Dict[str, List[Well]]:
        """
        Accessor function to navigate through a labware by row name.

        Use indexing to access individual rows or wells contained in the dictionary.
        For example, access row A with ``labware.rows_by_name()["A"]``.
        On a standard 96-well plate, this will output a list of :py:class:`.Well`
        objects containing A1 through A12.

        :return: Dictionary of :py:class:`.Well` lists keyed by row name.
        """
        return {
            row_name: [self._wells_by_name[well_name] for well_name in row]
            for row_name, row in self._well_grid.rows_by_name.items()
        }

    @requires_version(2, 0)
    def rows_by_index(self) -> Dict[str, List[Well]]:
        """
        .. deprecated:: 2.0
            Use :py:meth:`rows_by_name` instead.
        """
        _log.warning("rows_by_index is deprecated. Use rows_by_name instead.")
        return self.rows_by_name()

    @requires_version(2, 0)
    def columns(self, *args: Union[int, str]) -> List[List[Well]]:
        """
        Accessor function to navigate through a labware by column.

        Use indexing to access individual columns or wells contained in the nested list.
        For example, access column 1 with ``labware.columns()[0]``.
        On a standard 96-well plate, this will output a list of :py:class:`.Well`
        objects containing A1 through H1.

        .. note::
            Using args with this method is deprecated. Use indexing instead.

            If your code uses args, they can be either strings or integers, but not a
            mix of the two. For example, ``.columns(1, 4)`` or ``.columns("1", "4")`` is
            valid, but ``.columns("1", 4)`` is not.

        :return: A list of column lists.
        """
        if not args:
            return [
                [self._wells_by_name[well_name] for well_name in column]
                for column in self._well_grid.columns_by_name.values()
            ]

        elif validation.is_all_integers(args):
            columns = self.columns()
            return [columns[idx] for idx in args]

        elif validation.is_all_strings(args):
            columns_by_name = self.columns_by_name()
            return [columns_by_name[idx] for idx in args]

        else:
            raise TypeError(
                "`Labware.columns` must be called with all `int`'s or all `str`'s,"
                f" but was called with {args}"
            )

    @requires_version(2, 0)
    def columns_by_name(self) -> Dict[str, List[Well]]:
        """
        Accessor function to navigate through a labware by column name.

        Use indexing to access individual columns or wells contained in the dictionary.
        For example, access column 1 with ``labware.columns_by_name()["1"]``.
        On a standard 96-well plate, this will output a list of :py:class:`.Well`
        objects containing A1 through H1.

        :return: Dictionary of :py:class:`.Well` lists keyed by column name.
        """
        return {
            column_name: [self._wells_by_name[well_name] for well_name in column]
            for column_name, column in self._well_grid.columns_by_name.items()
        }

    @requires_version(2, 0)
    def columns_by_index(self) -> Dict[str, List[Well]]:
        """
        .. deprecated:: 2.0
            Use :py:meth:`columns_by_name` instead.
        """
        _log.warning("columns_by_index is deprecated. Use columns_by_name instead.")
        return self.columns_by_name()

    @property
    @requires_version(2, 0)
    def highest_z(self) -> float:
        """
        The z-coordinate of the highest single point anywhere on the labware.

        This is taken from the ``zDimension`` property of the ``dimensions`` object in the
        labware definition and takes into account the labware offset.
        """
        return self._core.highest_z

    @property
    def _is_tiprack(self) -> bool:
        """as is_tiprack but not subject to version checking for speed"""
        return self._core.is_tip_rack()

    @property
    @requires_version(2, 0)
    def is_tiprack(self) -> bool:
        """Whether the labware behaves as a tip rack.

        Returns ``True`` if the labware definition specifies ``isTiprack`` as ``True``.
        """
        return self._is_tiprack

    @property
    @requires_version(2, 15)
    def is_adapter(self) -> bool:
        """Whether the labware behaves as an adapter.

        Returns ``True`` if the labware definition specifies ``adapter`` as one of the
        labware's ``allowedRoles``.
        """
        return self._core.is_adapter()

    @property
    @requires_version(2, 0)
    def tip_length(self) -> float:
        """For a tip rack labware, the length of the tips it holds, in mm.

        This is taken from the ``tipLength`` property of the ``parameters`` object in the labware definition.

        This method will raise an exception if you call it on a labware that isn’t a tip rack.
        """
        return self._core.get_tip_length()

    @tip_length.setter
    def tip_length(self, length: float) -> None:
        """
        Set the tip rack's tip length.

        .. deprecated: 2.14
            Ensure tip length is set properly in your tip rack's definition
            and/or use the Opentrons App's tip length calibration feature.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.tip_length setter",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
            )

        # TODO(mc, 2023-02-06): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyLabwareCore)
        cast(LegacyLabwareCore, self._core).set_tip_length(length)

    # TODO(mc, 2022-11-09): implementation detail; deprecate public method
    def next_tip(
        self,
        num_tips: int = 1,
        starting_tip: Optional[Well] = None,
        *,
        nozzle_map: Optional[NozzleMapInterface] = None,
    ) -> Optional[Well]:
        """
        Find the next valid well for pick-up.

        Determines the next valid start tip from which to retrieve the
        specified number of tips. There must be at least `num_tips` sequential
        wells for which all wells have tips, in the same column.

        :param num_tips: target number of sequential tips in the same column
        :type num_tips: int
        :param starting_tip: The :py:class:`.Well` from which to start search.
                for an available tip.
        :type starting_tip: :py:class:`.Well`
        :return: the :py:class:`.Well` meeting the target criteria, or None
        """
        assert num_tips > 0, f"num_tips must be positive integer, but got {num_tips}"

        well_name = self._core.get_next_tip(
            num_tips=num_tips,
            starting_tip=starting_tip._core if starting_tip else None,
            nozzle_map=nozzle_map,
        )

        return self._wells_by_name[well_name] if well_name is not None else None

    def use_tips(self, start_well: Well, num_channels: int = 1) -> None:
        """
        Removes tips from the tip tracker.

        This method should be called when a tip is picked up. Generally, it
        will be called with `num_channels=1` or `num_channels=8` for single-
        and multi-channel respectively. If picking up with more than one
        channel, this method will automatically determine which tips are used
        based on the start well, the number of channels, and the geometry of
        the tiprack.

        :param start_well: The :py:class:`.Well` from which to pick up a tip.
                           For a single-channel pipette, this is the well to
                           send the pipette to. For a multi-channel pipette,
                           this is the well to send the back-most nozzle of the
                           pipette to.
        :type start_well: :py:class:`.Well`
        :param num_channels: The number of channels for the current pipette
        :type num_channels: int

        .. deprecated:: 2.14
            Modification of tip tracking state outside :py:meth:`.reset` has been deprecated.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.use_tips",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
                extra_message="To modify tip state, use Labware.reset.",
            )

        assert num_channels > 0, "Bad call to use_tips: num_channels<=0"
        fail_if_full = self._api_version < APIVersion(2, 2)

        # TODO(mc, 2023-02-13): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyLabwareCore)
        cast(LegacyLabwareCore, self._core).get_tip_tracker().use_tips(
            start_well=start_well._core,
            num_channels=num_channels,
            fail_if_full=fail_if_full,
        )

    def __repr__(self) -> str:
        return self._core.get_display_name()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Labware):
            return NotImplemented
        return self._core == other._core

    def __hash__(self) -> int:
        return hash((self._core, self._api_version))

    def previous_tip(self, num_tips: int = 1) -> Optional[Well]:
        """
        Find the best well to drop a tip in.

        This is the well from which the last tip was picked up, if there's
        room. It can be used to return tips to the tip tracker.

        :param num_tips: target number of tips to return, sequential in a
                         column
        :type num_tips: int
        :return: The :py:class:`.Well` meeting the target criteria, or ``None``

        .. versionchanged:: 2.14
            This method has been removed.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.previous_tip",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
            )

        # This logic is the inverse of :py:meth:`next_tip`
        assert num_tips > 0, "Bad call to previous_tip: num_tips <= 0"
        # TODO(mc, 2023-02-13): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyLabwareCore)
        well_core = (
            cast(LegacyLabwareCore, self._core)
            .get_tip_tracker()
            .previous_tip(num_tips=num_tips)
        )
        return self._wells_by_name[well_core.get_name()] if well_core else None

    # TODO(mc, 2022-11-09): implementation detail; deprecate public method
    def return_tips(self, start_well: Well, num_channels: int = 1) -> None:
        """
        Re-adds tips to the tip tracker

        This method should be called when a tip is dropped in a tiprack. It
        should be called with ``num_channels=1`` or ``num_channels=8`` for
        single- and multi-channel respectively. If returning more than one
        channel, this method will automatically determine which tips are
        returned based on the start well, the number of channels,
        and the tiprack geometry.

        Note that unlike :py:meth:`use_tips`, calling this method in a way
        that would drop tips into wells with tips in them will raise an
        exception; this should only be called on a valid return of
        :py:meth:`previous_tip`.

        :param start_well: The :py:class:`.Well` into which to return a tip.
        :type start_well: :py:class:`.Well`
        :param num_channels: The number of channels for the current pipette
        :type num_channels: int

        .. versionchanged:: 2.14
            This method has been removed. Use :py:meth:`.reset` instead.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise UnsupportedAPIError(
                api_element="Labware.return_tips()",
                since_version=f"{ENGINE_CORE_API_VERSION}",
                current_version=f"{self._api_version}",
                extra_message="Use Labware.reset() instead.",
            )

        # This logic is the inverse of :py:meth:`use_tips`
        assert num_channels > 0, "Bad call to return_tips: num_channels <= 0"

        # TODO(mc, 2023-02-13): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyLabwareCore)
        cast(LegacyLabwareCore, self._core).get_tip_tracker().return_tips(
            start_well=start_well._core, num_channels=num_channels
        )

    @requires_version(2, 0)
    def reset(self) -> None:
        """Reset tip tracking for a tip rack.

        After resetting, the API treats all wells on the rack as if they contain unused tips.
        This is useful if you want to reuse tips after calling :py:meth:`.return_tip()`.

        If you need to physically replace an empty tip rack in the middle of your protocol,
        use :py:meth:`.move_labware()` instead. See :ref:`off-deck-location` for an example.

        .. versionchanged:: 2.14
            This method will raise an exception if you call it on a labware that isn't
            a tip rack. Formerly, it would do nothing.
        """
        self._core.reset_tips()

    @requires_version(2, 22)
    def load_liquid(
        self, wells: Sequence[Union[str, Well]], volume: float, liquid: Liquid
    ) -> None:
        """Mark several wells as containing the same amount of liquid.

        This method should be called at the beginning of a protocol, soon after loading labware and before
        liquid handling operations begin. Loading liquids is required for liquid tracking functionality. If a well
        hasn't been assigned a starting volume with :py:meth:`~Labware.load_empty`, :py:meth:`~Labware.load_liquid`, or
        :py:meth:`~Labware.load_liquid_by_well`, the volume it contains is unknown and the well's liquid will not be tracked throughout the protocol.

        :param wells: The wells to load the liquid into.
        :type wells: List of string well names or list of :py:class:`.Well` objects (e.g., from :py:meth:`~Labware.wells`).

        :param volume: The volume of liquid to load into each well.
        :type volume: float

        :param liquid: The liquid to load into each well, previously defined by :py:meth:`~ProtocolContext.define_liquid`
        :type liquid: Liquid
        """
        well_names: List[str] = []
        for well in wells:
            if isinstance(well, str):
                if well not in self.wells_by_name():
                    raise KeyError(
                        f"{well} is not a well in labware {self.name}. The elements of wells should name wells in this labware."
                    )
                well_names.append(well)
            elif isinstance(well, Well):
                if well.parent is not self:
                    raise KeyError(
                        f"{well.well_name} is not a well in labware {self.name}. The elements of wells should be wells of this labware."
                    )
                well_names.append(well.well_name)
            else:
                raise TypeError(
                    f"Unexpected type for element {repr(well)}. The elements of wells should be Well instances or well names."
                )
            if not isinstance(volume, (float, int)):
                raise TypeError(
                    f"Unexpected type for volume {repr(volume)}. Volume should be a number in microliters."
                )
        self._core.load_liquid({well_name: volume for well_name in well_names}, liquid)

    @requires_version(2, 22)
    def load_liquid_by_well(
        self, volumes: Mapping[Union[str, Well], float], liquid: Liquid
    ) -> None:
        """Mark several wells as containing unique volumes of liquid.

        This method should be called at the beginning of a protocol, soon after loading labware and before
        liquid handling begins. Loading liquids is required for liquid tracking functionality. If a well hasn't been assigned a starting volume with :py:meth:`~Labware.load_empty`, :py:meth:`~Labware.load_liquid`, or
        :py:meth:`~Labware.load_liquid_by_well`, the volume it contains is unknown and the well's liquid will not be tracked throughout the protocol.

        :param volumes: A dictionary of well names (or :py:class:`Well` objects, for instance from ``labware['A1']``)
        :type wells: Dict[Union[str, Well], float]

        :param liquid: The liquid to load into each well, previously defined by :py:meth:`~ProtocolContext.define_liquid`
        :type liquid: Liquid
        """
        verified_volumes: Dict[str, float] = {}
        for well, volume in volumes.items():
            if isinstance(well, str):
                if well not in self.wells_by_name():
                    raise KeyError(
                        f"{well} is not a well in {self.name}. The keys of volumes should name wells in this labware"
                    )
                verified_volumes[well] = volume
            elif isinstance(well, Well):
                if well.parent is not self:
                    raise KeyError(
                        f"{well.well_name} is not a well in {self.name}. The keys of volumes should be wells of this labware"
                    )
                verified_volumes[well.well_name] = volume
            else:
                raise TypeError(
                    f"Unexpected type for well name {repr(well)}. The keys of volumes should be Well instances or well names."
                )
            if not isinstance(volume, (float, int)):
                raise TypeError(
                    f"Unexpected type for volume {repr(volume)}. The values of volumes should be numbers in microliters."
                )
        self._core.load_liquid(verified_volumes, liquid)

    @requires_version(2, 22)
    def load_empty(self, wells: Sequence[Union[Well, str]]) -> None:
        """Mark several wells as empty.

        This method should be called at the beginning of a protocol, after loading the labware and before liquid handling
        begins. Loading liquids is required for liquid tracking functionality. If a well in a labware hasn't been assigned a starting volume with :py:meth:`Labware.load_empty`, :py:meth:`Labware.load_liquid`, or :py:meth:`Labware.load_liquid_by_well`, the
        volume it contains is unknown and the well's liquid will not be tracked throughout the protocol.

        :param wells: The list of wells to mark empty. To mark all wells as empty, pass ``labware.wells()``. You can also specify
                      wells by their names (for instance, ``labware.load_empty(['A1', 'A2'])``).
        :type wells: Union[List[Well], List[str]]
        """
        well_names: List[str] = []
        for well in wells:
            if isinstance(well, str):
                if well not in self.wells_by_name():
                    raise KeyError(
                        f"{well} is not a well in {self.name}. The elements of wells should name wells in this labware."
                    )
                well_names.append(well)
            elif isinstance(well, Well):
                if well.parent is not self:
                    raise KeyError(
                        f"{well.well_name} is not a well in {self.name}. The elements of wells should be wells of this labware."
                    )
                well_names.append(well.well_name)
            else:
                raise TypeError(
                    f"Unexpected type for well name {repr(well)}. The elements of wells should be Well instances or well names."
                )
        self._core.load_empty(well_names)


# TODO(mc, 2022-11-09): implementation detail, move to core
def split_tipracks(tip_racks: List[Labware]) -> Tuple[Labware, List[Labware]]:
    try:
        rest = tip_racks[1:]
    except IndexError:
        rest = []
    return tip_racks[0], rest


# TODO(mc, 2022-11-09): implementation detail, move to core
def select_tiprack_from_list(
    tip_racks: List[Labware],
    num_channels: int,
    starting_point: Optional[Well] = None,
    *,
    nozzle_map: Optional[NozzleMapInterface] = None,
) -> Tuple[Labware, Well]:
    try:
        first, rest = split_tipracks(tip_racks)
    except IndexError:
        raise OutOfTipsError

    if starting_point and starting_point.parent != first:
        raise TipSelectionError(
            f"The starting tip you selected does not exist in {first}"
        )
    elif starting_point:
        first_well = starting_point
    elif nozzle_map:
        first_well = None
    else:
        first_well = first.wells()[0]

    next_tip = first.next_tip(num_channels, first_well, nozzle_map=nozzle_map)
    if next_tip:
        return first, next_tip
    else:
        return select_tiprack_from_list(rest, num_channels, None, nozzle_map=nozzle_map)


# TODO(mc, 2022-11-09): implementation detail, move to core
def filter_tipracks_to_start(
    starting_point: Well, tipracks: List[Labware]
) -> List[Labware]:
    return list(dropwhile(lambda tr: starting_point.parent != tr, tipracks))


# TODO(mc, 2022-11-09): implementation detail, move to core
def next_available_tip(
    starting_tip: Optional[Well],
    tip_racks: List[Labware],
    channels: int,
    *,
    nozzle_map: Optional[NozzleMapInterface] = None,
) -> Tuple[Labware, Well]:
    start = starting_tip
    if start is None:
        return select_tiprack_from_list(
            tip_racks, channels, None, nozzle_map=nozzle_map
        )
    else:
        return select_tiprack_from_list(
            filter_tipracks_to_start(start, tip_racks),
            channels,
            start,
            nozzle_map=nozzle_map,
        )


# TODO(mc, 2022-11-09): implementation detail, move somewhere else
# only used in old calibration flows by robot-server
def load_from_definition(
    definition: "LabwareDefinition2",
    parent: Location,
    label: Optional[str] = None,
    api_level: Optional[APIVersion] = None,
) -> Labware:
    """
    Return a labware object constructed from a provided labware definition dict

    :param definition: A dict representing all required data for a labware,
        including metadata such as the display name of the labware, a
        definition of the order to iterate over wells, the shape of wells
        (shape, physical dimensions, etc), and so on. The correct shape of
        this definition is governed by the "labware-designer" project in
        the Opentrons/opentrons repo.
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of labware is
                   (often the front-left corner of a slot on the deck).
    :param str label: An optional label that will override the labware's
                      display name from its definition
    :param api_level: the API version to set for the loaded labware
                      instance. The :py:class:`.Labware` will
                      conform to this level. If not specified,
                      defaults to ``APIVersion(2, 13)``.
    """
    return Labware(
        core=LegacyLabwareCore(
            definition=definition,
            parent=parent,
            label=label,
        ),
        api_version=api_level or APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )


# TODO(mc, 2022-11-09): implementation detail, move somewhere else
# only used in old calibration flows by robot-server
def load(
    load_name: str,
    parent: Location,
    label: Optional[str] = None,
    namespace: Optional[str] = None,
    version: int = 1,
    bundled_defs: Optional[Mapping[str, LabwareDefinition2]] = None,
    extra_defs: Optional[Mapping[str, LabwareDefinition2]] = None,
    api_level: Optional[APIVersion] = None,
) -> Labware:
    """
    Return a labware object constructed from a labware definition dict looked
    up by name (definition must have been previously stored locally on the
    robot)

    :param load_name: A string to use for looking up a labware definition
        previously saved to disc. The definition file must have been saved in a
        known location
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of labware is
                   (often the front-left corner of a slot on the deck).
    :param str label: An optional label that will override the labware's
                      display name from its definition
    :param str namespace: The namespace the labware definition belongs to.
        If unspecified, will search 'opentrons' then 'custom_beta'
    :param int version: The version of the labware definition. If unspecified,
        will use version 1.
    :param bundled_defs: If specified, a mapping of labware names to labware
        definitions. Only the bundle will be searched for definitions.
    :param extra_defs: If specified, a mapping of labware names to labware
        definitions. If no bundle is passed, these definitions will also be
        searched.
    :param api_level: the API version to set for the loaded labware
                      instance. The :py:class:`.Labware` will
                      conform to this level. If not specified,
                      defaults to ``APIVersion(2, 13)``.
    """
    definition = get_labware_definition(
        load_name,
        namespace,
        version,
        bundled_defs=bundled_defs,
        extra_defs=extra_defs,
    )

    # The legacy `load_from_definition()` function that we're calling only supports
    # schemaVersion==2 labware. Fortunately, when robot-server calls this function,
    # we only expect it to try to load schemaVersion==2 labware, so we never expect
    # this ValueError to be raised in practice.
    if definition["schemaVersion"] != 2:
        raise ValueError(
            f"{namespace}/{load_name}/{version} has schema {definition['schemaVersion']}."
            " Only schema 2 is supported."
        )

    return load_from_definition(definition, parent, label, api_level)



"""Protocol engine command models.

The `protocol_engine.commands` module contains the data models of all
command requests and results the engine is able to handle. It also contains
models and type definitions for command resources, where a command:

- May be pending, running, completed, or failed
- Contains request and, if applicable, result data
- Has timestamps for when it was created, started, completed, and/or failed

Request and result models in the module are defined using Pydantic, because
they are part of the public input / output of the engine, and need validation
and/or schema generation.
"""

from . import absorbance_reader
from . import flex_stacker
from . import heater_shaker
from . import magnetic_module
from . import temperature_module
from . import thermocycler
from . import calibration
from . import unsafe
from . import robot

from .hash_command_params import hash_protocol_command_params
from .generate_command_schema import generate_command_schema

from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    CommandStatus,
    CommandIntent,
)

from .command_unions import (
    Command,
    CommandAdapter,
    CommandParams,
    CommandCreate,
    CommandCreateAdapter,
    CommandResult,
    CommandType,
    CommandDefinedErrorData,
)

from .air_gap_in_place import (
    AirGapInPlace,
    AirGapInPlaceParams,
    AirGapInPlaceCreate,
    AirGapInPlaceResult,
    AirGapInPlaceCommandType,
)

from .aspirate import (
    Aspirate,
    AspirateParams,
    AspirateCreate,
    AspirateResult,
    AspirateCommandType,
)

from .aspirate_while_tracking import (
    AspirateWhileTracking,
    AspirateWhileTrackingParams,
    AspirateWhileTrackingCreate,
    AspirateWhileTrackingResult,
    AspirateWhileTrackingCommandType,
)

from .aspirate_in_place import (
    AspirateInPlace,
    AspirateInPlaceParams,
    AspirateInPlaceCreate,
    AspirateInPlaceResult,
    AspirateInPlaceCommandType,
)

from .comment import (
    Comment,
    CommentParams,
    CommentCreate,
    CommentResult,
    CommentCommandType,
)

from .custom import (
    Custom,
    CustomParams,
    CustomCreate,
    CustomResult,
    CustomCommandType,
)

from .dispense import (
    Dispense,
    DispenseParams,
    DispenseCreate,
    DispenseResult,
    DispenseCommandType,
)

from .dispense_while_tracking import (
    DispenseWhileTracking,
    DispenseWhileTrackingParams,
    DispenseWhileTrackingCreate,
    DispenseWhileTrackingResult,
    DispenseWhileTrackingCommandType,
)

from .dispense_in_place import (
    DispenseInPlace,
    DispenseInPlaceParams,
    DispenseInPlaceCreate,
    DispenseInPlaceResult,
    DispenseInPlaceCommandType,
)

from .drop_tip import (
    DropTip,
    DropTipParams,
    DropTipCreate,
    DropTipResult,
    DropTipCommandType,
)

from .drop_tip_in_place import (
    DropTipInPlace,
    DropTipInPlaceParams,
    DropTipInPlaceCreate,
    DropTipInPlaceResult,
    DropTipInPlaceCommandType,
)

from .home import (
    Home,
    HomeParams,
    HomeCreate,
    HomeResult,
    HomeCommandType,
)

from .load_labware import (
    LoadLabware,
    LoadLabwareParams,
    LoadLabwareCreate,
    LoadLabwareResult,
    LoadLabwareCommandType,
)

from .reload_labware import (
    ReloadLabware,
    ReloadLabwareParams,
    ReloadLabwareCreate,
    ReloadLabwareResult,
    ReloadLabwareCommandType,
)

from .load_liquid import (
    LoadLiquid,
    LoadLiquidParams,
    LoadLiquidResult,
    LoadLiquidCreate,
    LoadLiquidCommandType,
    LoadLiquidImplementation,
)

from .load_liquid_class import (
    LoadLiquidClass,
    LoadLiquidClassParams,
    LoadLiquidClassCreate,
    LoadLiquidClassResult,
    LoadLiquidClassCommandType,
    LoadLiquidClassImplementation,
)

from .load_module import (
    LoadModule,
    LoadModuleParams,
    LoadModuleCreate,
    LoadModuleResult,
    LoadModuleCommandType,
)

from .load_pipette import (
    LoadPipette,
    LoadPipetteParams,
    LoadPipetteCreate,
    LoadPipetteResult,
    LoadPipetteCommandType,
)

from .load_lid_stack import (
    LoadLidStack,
    LoadLidStackParams,
    LoadLidStackCreate,
    LoadLidStackResult,
    LoadLidStackCommandType,
)

from .load_lid import (
    LoadLid,
    LoadLidParams,
    LoadLidCreate,
    LoadLidResult,
    LoadLidCommandType,
)

from .move_labware import (
    MoveLabware,
    MoveLabwareParams,
    MoveLabwareCreate,
    MoveLabwareResult,
    MoveLabwareCommandType,
)

from .move_relative import (
    MoveRelative,
    MoveRelativeParams,
    MoveRelativeCreate,
    MoveRelativeResult,
    MoveRelativeCommandType,
)

from .move_to_coordinates import (
    MoveToCoordinates,
    MoveToCoordinatesParams,
    MoveToCoordinatesCreate,
    MoveToCoordinatesResult,
    MoveToCoordinatesCommandType,
)

from .move_to_well import (
    MoveToWell,
    MoveToWellParams,
    MoveToWellCreate,
    MoveToWellResult,
    MoveToWellCommandType,
)

from .move_to_addressable_area import (
    MoveToAddressableArea,
    MoveToAddressableAreaParams,
    MoveToAddressableAreaCreate,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaCommandType,
)

from .move_to_addressable_area_for_drop_tip import (
    MoveToAddressableAreaForDropTip,
    MoveToAddressableAreaForDropTipParams,
    MoveToAddressableAreaForDropTipCreate,
    MoveToAddressableAreaForDropTipResult,
    MoveToAddressableAreaForDropTipCommandType,
)

from .wait_for_resume import (
    WaitForResume,
    WaitForResumeParams,
    WaitForResumeCreate,
    WaitForResumeResult,
    WaitForResumeCommandType,
)

from .wait_for_duration import (
    WaitForDuration,
    WaitForDurationParams,
    WaitForDurationCreate,
    WaitForDurationResult,
    WaitForDurationCommandType,
)

from .pick_up_tip import (
    PickUpTip,
    PickUpTipParams,
    PickUpTipCreate,
    PickUpTipResult,
    PickUpTipCommandType,
)

from .save_position import (
    SavePosition,
    SavePositionParams,
    SavePositionCreate,
    SavePositionResult,
    SavePositionCommandType,
)

from .set_rail_lights import (
    SetRailLights,
    SetRailLightsParams,
    SetRailLightsCreate,
    SetRailLightsResult,
    SetRailLightsCommandType,
)

from .touch_tip import (
    TouchTip,
    TouchTipParams,
    TouchTipCreate,
    TouchTipResult,
    TouchTipCommandType,
)

from .blow_out import (
    BlowOutParams,
    BlowOutResult,
    BlowOutCreate,
    BlowOutImplementation,
    BlowOut,
)

from .blow_out_in_place import (
    BlowOutInPlaceParams,
    BlowOutInPlaceResult,
    BlowOutInPlaceCreate,
    BlowOutInPlaceImplementation,
    BlowOutInPlace,
)

from .set_status_bar import (
    SetStatusBar,
    SetStatusBarParams,
    SetStatusBarCreate,
    SetStatusBarResult,
    SetStatusBarImplementation,
    SetStatusBarCommandType,
)

from .retract_axis import (
    RetractAxis,
    RetractAxisParams,
    RetractAxisCreate,
    RetractAxisResult,
    RetractAxisCommandType,
)

from .configure_for_volume import (
    ConfigureForVolume,
    ConfigureForVolumeCreate,
    ConfigureForVolumeParams,
    ConfigureForVolumeResult,
    ConfigureForVolumeCommandType,
)

from .prepare_to_aspirate import (
    PrepareToAspirate,
    PrepareToAspirateCreate,
    PrepareToAspirateParams,
    PrepareToAspirateResult,
    PrepareToAspirateCommandType,
)

from .configure_nozzle_layout import (
    ConfigureNozzleLayout,
    ConfigureNozzleLayoutCreate,
    ConfigureNozzleLayoutParams,
    ConfigureNozzleLayoutResult,
    ConfigureNozzleLayoutCommandType,
)

from .get_tip_presence import (
    GetTipPresence,
    GetTipPresenceCreate,
    GetTipPresenceParams,
    GetTipPresenceResult,
    GetTipPresenceCommandType,
)

from .verify_tip_presence import (
    VerifyTipPresence,
    VerifyTipPresenceCreate,
    VerifyTipPresenceParams,
    VerifyTipPresenceResult,
    VerifyTipPresenceCommandType,
)

from .get_next_tip import (
    GetNextTip,
    GetNextTipCreate,
    GetNextTipParams,
    GetNextTipResult,
    GetNextTipCommandType,
)

from .liquid_probe import (
    LiquidProbe,
    LiquidProbeParams,
    LiquidProbeCreate,
    LiquidProbeResult,
    LiquidProbeCommandType,
    TryLiquidProbe,
    TryLiquidProbeParams,
    TryLiquidProbeCreate,
    TryLiquidProbeResult,
    TryLiquidProbeCommandType,
)

from .seal_pipette_to_tip import (
    SealPipetteToTip,
    SealPipetteToTipParams,
    SealPipetteToTipCreate,
    SealPipetteToTipResult,
    SealPipetteToTipCommandType,
)
from .unseal_pipette_from_tip import (
    UnsealPipetteFromTip,
    UnsealPipetteFromTipParams,
    UnsealPipetteFromTipCreate,
    UnsealPipetteFromTipResult,
    UnsealPipetteFromTipCommandType,
)
from .pressure_dispense import (
    PressureDispense,
    PressureDispenseParams,
    PressureDispenseCreate,
    PressureDispenseResult,
    PressureDispenseCommandType,
)

__all__ = [
    # command type unions
    "Command",
    "CommandAdapter",
    "CommandParams",
    "CommandCreate",
    "CommandCreateAdapter",
    "CommandResult",
    "CommandType",
    "CommandPrivateResult",
    "CommandDefinedErrorData",
    # base interfaces
    "AbstractCommandImpl",
    "BaseCommand",
    "BaseCommandCreate",
    "CommandStatus",
    "CommandIntent",
    # command parameter hashing
    "hash_protocol_command_params",
    # command schema generation
    "generate_command_schema",
    # air gap command models
    "AirGapInPlace",
    "AirGapInPlaceCreate",
    "AirGapInPlaceParams",
    "AirGapInPlaceResult",
    "AirGapInPlaceCommandType",
    # aspirate command models
    "Aspirate",
    "AspirateCreate",
    "AspirateParams",
    "AspirateResult",
    "AspirateCommandType",
    # aspirate while tracking command models
    "AspirateWhileTracking",
    "AspirateWhileTrackingCreate",
    "AspirateWhileTrackingParams",
    "AspirateWhileTrackingResult",
    "AspirateWhileTrackingCommandType",
    # aspirate in place command models
    "AspirateInPlace",
    "AspirateInPlaceCreate",
    "AspirateInPlaceParams",
    "AspirateInPlaceResult",
    "AspirateInPlaceCommandType",
    # comment command models
    "Comment",
    "CommentParams",
    "CommentCreate",
    "CommentResult",
    "CommentCommandType",
    # custom command models
    "Custom",
    "CustomCreate",
    "CustomParams",
    "CustomResult",
    "CustomCommandType",
    # dispense command models
    "Dispense",
    "DispenseCreate",
    "DispenseParams",
    "DispenseResult",
    "DispenseCommandType",
    # dispense while tracking command models
    "DispenseWhileTracking",
    "DispenseWhileTrackingCreate",
    "DispenseWhileTrackingParams",
    "DispenseWhileTrackingResult",
    "DispenseWhileTrackingCommandType",
    # dispense in place command models
    "DispenseInPlace",
    "DispenseInPlaceCreate",
    "DispenseInPlaceParams",
    "DispenseInPlaceResult",
    "DispenseInPlaceCommandType",
    # drop tip command models
    "DropTip",
    "DropTipCreate",
    "DropTipParams",
    "DropTipResult",
    "DropTipCommandType",
    # drop tip in place command models
    "DropTipInPlace",
    "DropTipInPlaceCreate",
    "DropTipInPlaceParams",
    "DropTipInPlaceResult",
    "DropTipInPlaceCommandType",
    # home command models
    "Home",
    "HomeParams",
    "HomeCreate",
    "HomeResult",
    "HomeCommandType",
    # retract axis command models
    "RetractAxis",
    "RetractAxisCreate",
    "RetractAxisParams",
    "RetractAxisResult",
    "RetractAxisCommandType",
    # load labware command models
    "LoadLabware",
    "LoadLabwareCreate",
    "LoadLabwareParams",
    "LoadLabwareResult",
    "LoadLabwareCommandType",
    # reload labware command models
    "ReloadLabware",
    "ReloadLabwareCreate",
    "ReloadLabwareParams",
    "ReloadLabwareResult",
    "ReloadLabwareCommandType",
    # load module command models
    "LoadModule",
    "LoadModuleCreate",
    "LoadModuleParams",
    "LoadModuleResult",
    "LoadModuleCommandType",
    # load pipette command models
    "LoadPipette",
    "LoadPipetteCreate",
    "LoadPipetteParams",
    "LoadPipetteResult",
    "LoadPipetteCommandType",
    "LoadPipettePrivateResult",
    # load lid stack command models
    "LoadLidStack",
    "LoadLidStackCreate",
    "LoadLidStackParams",
    "LoadLidStackResult",
    "LoadLidStackCommandType",
    "LoadLidStackPrivateResult",
    # load lid command models
    "LoadLid",
    "LoadLidCreate",
    "LoadLidParams",
    "LoadLidResult",
    "LoadLidCommandType",
    "LoadLidPrivateResult",
    # move labware command models
    "MoveLabware",
    "MoveLabwareCreate",
    "MoveLabwareParams",
    "MoveLabwareResult",
    "MoveLabwareCommandType",
    # move relative command models
    "MoveRelative",
    "MoveRelativeParams",
    "MoveRelativeCreate",
    "MoveRelativeResult",
    "MoveRelativeCommandType",
    # move to coordinates command models
    "MoveToCoordinates",
    "MoveToCoordinatesParams",
    "MoveToCoordinatesCreate",
    "MoveToCoordinatesResult",
    "MoveToCoordinatesCommandType",
    # move to well command models
    "MoveToWell",
    "MoveToWellCreate",
    "MoveToWellParams",
    "MoveToWellResult",
    "MoveToWellCommandType",
    # move to addressable area command models
    "MoveToAddressableArea",
    "MoveToAddressableAreaParams",
    "MoveToAddressableAreaCreate",
    "MoveToAddressableAreaResult",
    "MoveToAddressableAreaCommandType",
    # move to addressable area for drop tip command models
    "MoveToAddressableAreaForDropTip",
    "MoveToAddressableAreaForDropTipParams",
    "MoveToAddressableAreaForDropTipCreate",
    "MoveToAddressableAreaForDropTipResult",
    "MoveToAddressableAreaForDropTipCommandType",
    # wait for resume command models
    "WaitForResume",
    "WaitForResumeParams",
    "WaitForResumeCreate",
    "WaitForResumeResult",
    "WaitForResumeCommandType",
    # wait for duration command models
    "WaitForDuration",
    "WaitForDurationParams",
    "WaitForDurationCreate",
    "WaitForDurationResult",
    "WaitForDurationCommandType",
    # pick up tip command models
    "PickUpTip",
    "PickUpTipCreate",
    "PickUpTipParams",
    "PickUpTipResult",
    "PickUpTipCommandType",
    # save position command models
    "SavePosition",
    "SavePositionParams",
    "SavePositionCreate",
    "SavePositionResult",
    "SavePositionCommandType",
    # set rail lights command models
    "SetRailLights",
    "SetRailLightsParams",
    "SetRailLightsCreate",
    "SetRailLightsResult",
    "SetRailLightsCommandType",
    # touch tip command models
    "TouchTip",
    "TouchTipParams",
    "TouchTipCreate",
    "TouchTipResult",
    "TouchTipCommandType",
    # blow out command models
    "BlowOutResult",
    "BlowOutCreate",
    "BlowOutImplementation",
    "BlowOutParams",
    "BlowOut",
    # blow out in place command models
    "BlowOutInPlaceParams",
    "BlowOutInPlaceResult",
    "BlowOutInPlaceCreate",
    "BlowOutInPlaceImplementation",
    "BlowOutInPlace",
    # set status bar command models
    "SetStatusBar",
    "SetStatusBarParams",
    "SetStatusBarCreate",
    "SetStatusBarResult",
    "SetStatusBarImplementation",
    "SetStatusBarCommandType",
    # load liquid command models
    "LoadLiquid",
    "LoadLiquidCreate",
    "LoadLiquidImplementation",
    "LoadLiquidParams",
    "LoadLiquidResult",
    "LoadLiquidCommandType",
    # load liquid class command models
    "LoadLiquidClass",
    "LoadLiquidClassParams",
    "LoadLiquidClassCreate",
    "LoadLiquidClassResult",
    "LoadLiquidClassImplementation",
    "LoadLiquidClassCommandType",
    # hardware control command models
    # hardware module command bundles
    "absorbance_reader",
    "flex_stacker",
    "heater_shaker",
    "magnetic_module",
    "temperature_module",
    "thermocycler",
    # calibration command bundle
    "calibration",
    # unsafe command bundle
    "unsafe",
    "robot",
    # configure pipette volume command bundle
    "ConfigureForVolume",
    "ConfigureForVolumeCreate",
    "ConfigureForVolumeParams",
    "ConfigureForVolumeResult",
    "ConfigureForVolumeCommandType",
    # prepare pipette for aspirate command bundle
    "PrepareToAspirate",
    "PrepareToAspirateCreate",
    "PrepareToAspirateParams",
    "PrepareToAspirateResult",
    "PrepareToAspirateCommandType",
    # configure nozzle layout command bundle
    "ConfigureNozzleLayout",
    "ConfigureNozzleLayoutCreate",
    "ConfigureNozzleLayoutParams",
    "ConfigureNozzleLayoutResult",
    "ConfigureNozzleLayoutCommandType",
    # get pipette tip presence bundle
    "GetTipPresence",
    "GetTipPresenceCreate",
    "GetTipPresenceParams",
    "GetTipPresenceResult",
    "GetTipPresenceCommandType",
    # verify pipette tip presence bundle
    "VerifyTipPresence",
    "VerifyTipPresenceCreate",
    "VerifyTipPresenceParams",
    "VerifyTipPresenceResult",
    "VerifyTipPresenceCommandType",
    # get next tip command bundle
    "GetNextTip",
    "GetNextTipCreate",
    "GetNextTipParams",
    "GetNextTipResult",
    "GetNextTipCommandType",
    # liquid probe command bundle
    "LiquidProbe",
    "LiquidProbeParams",
    "LiquidProbeCreate",
    "LiquidProbeResult",
    "LiquidProbeCommandType",
    "TryLiquidProbe",
    "TryLiquidProbeParams",
    "TryLiquidProbeCreate",
    "TryLiquidProbeResult",
    "TryLiquidProbeCommandType",
    # seal command bundle
    "SealPipetteToTip",
    "SealPipetteToTipParams",
    "SealPipetteToTipCreate",
    "SealPipetteToTipResult",
    "SealPipetteToTipCommandType",
    # unseal command bundle
    "UnsealPipetteFromTip",
    "UnsealPipetteFromTipParams",
    "UnsealPipetteFromTipCreate",
    "UnsealPipetteFromTipResult",
    "UnsealPipetteFromTipCommandType",
    # pressure dispense command bundle
    "PressureDispense",
    "PressureDispenseParams",
    "PressureDispenseCreate",
    "PressureDispenseResult",
    "PressureDispenseCommandType",
]



"""Dispense-in-place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union, Any
from typing_extensions import Literal

from pydantic import Field
from pydantic.json_schema import SkipJsonSchema

from .pipetting_common import (
    PipetteIdMixin,
    DispenseVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    OverpressureError,
    dispense_in_place,
    DEFAULT_CORRECTION_VOLUME,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..state.update_types import CLEAR
from ..types import CurrentWell

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover
    from ..resources import ModelUtils
    from ..state.state import StateView


DispenseInPlaceCommandType = Literal["dispenseInPlace"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class DispenseInPlaceParams(PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin):
    """Payload required to dispense in place."""

    pushOut: float | SkipJsonSchema[None] = Field(
        None,
        description="push the plunger a small amount farther than necessary for accurate low-volume dispensing",
        json_schema_extra=_remove_default,
    )


class DispenseInPlaceResult(BaseLiquidHandlingResult):
    """Result data from the execution of a DispenseInPlace command."""

    pass


_ExecuteReturn = Union[
    SuccessData[DispenseInPlaceResult],
    DefinedErrorData[OverpressureError],
]


class DispenseInPlaceImplementation(
    AbstractCommandImpl[DispenseInPlaceParams, _ExecuteReturn]
):
    """DispenseInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        gantry_mover: GantryMover,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._gantry_mover = gantry_mover
        self._model_utils = model_utils

    async def execute(self, params: DispenseInPlaceParams) -> _ExecuteReturn:
        """Dispense without moving the pipette."""
        current_location = self._state_view.pipettes.get_current_location()
        current_position = await self._gantry_mover.get_position(params.pipetteId)
        result = await dispense_in_place(
            pipette_id=params.pipetteId,
            volume=params.volume,
            flow_rate=params.flowRate,
            push_out=params.pushOut,
            location_if_error={
                "retryLocation": (
                    current_position.x,
                    current_position.y,
                    current_position.z,
                )
            },
            pipetting=self._pipetting,
            model_utils=self._model_utils,
            correction_volume=params.correctionVolume or DEFAULT_CORRECTION_VOLUME,
        )
        if isinstance(result, DefinedErrorData):
            if (
                isinstance(current_location, CurrentWell)
                and current_location.pipette_id == params.pipetteId
            ):
                return DefinedErrorData(
                    public=result.public,
                    state_update=result.state_update.set_liquid_operated(
                        labware_id=current_location.labware_id,
                        well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                            current_location.labware_id,
                            current_location.well_name,
                            params.pipetteId,
                        ),
                        volume_added=CLEAR,
                    ),
                    state_update_if_false_positive=result.state_update_if_false_positive,
                )
            else:
                return result
        else:
            if (
                isinstance(current_location, CurrentWell)
                and current_location.pipette_id == params.pipetteId
            ):
                volume_added = (
                    self._state_view.pipettes.get_liquid_dispensed_by_ejecting_volume(
                        pipette_id=params.pipetteId, volume=result.public.volume
                    )
                )
                if volume_added is not None:
                    volume_added *= self._state_view.geometry.get_nozzles_per_well(
                        current_location.labware_id,
                        current_location.well_name,
                        params.pipetteId,
                    )
                return SuccessData(
                    public=DispenseInPlaceResult(volume=result.public.volume),
                    state_update=result.state_update.set_liquid_operated(
                        labware_id=current_location.labware_id,
                        well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                            current_location.labware_id,
                            current_location.well_name,
                            params.pipetteId,
                        ),
                        volume_added=volume_added
                        if volume_added is not None
                        else CLEAR,
                    ),
                )
            else:
                return SuccessData(
                    public=DispenseInPlaceResult(volume=result.public.volume),
                    state_update=result.state_update,
                )


class DispenseInPlace(
    BaseCommand[DispenseInPlaceParams, DispenseInPlaceResult, OverpressureError]
):
    """DispenseInPlace command model."""

    commandType: DispenseInPlaceCommandType = "dispenseInPlace"
    params: DispenseInPlaceParams
    result: Optional[DispenseInPlaceResult] = None

    _ImplementationCls: Type[
        DispenseInPlaceImplementation
    ] = DispenseInPlaceImplementation


class DispenseInPlaceCreate(BaseCommandCreate[DispenseInPlaceParams]):
    """DispenseInPlace command request model."""

    commandType: DispenseInPlaceCommandType = "dispenseInPlace"
    params: DispenseInPlaceParams

    _CommandCls: Type[DispenseInPlace] = DispenseInPlace



"""Protocol engine module.

The protocol_engine module contains the logic necessary to take a stream of
protocol commands, issued by some arbitrary protocol runner, and turn it into
protocol state and side-effects like robot movements.

The main interface is the `ProtocolEngine` class.
"""

from .protocol_engine import ProtocolEngine
from .errors import ProtocolEngineError, ErrorOccurrence
from .notes import CommandNote
from .commands import (
    Command,
    CommandParams,
    CommandCreate,
    CommandStatus,
    CommandType,
    CommandIntent,
)
from .state.state import State, StateView
from .state.state_summary import StateSummary
from .state.commands import CommandSlice, CommandErrorSlice, CommandPointer
from .state.config import Config
from .plugins import AbstractPlugin

from .types import (
    LabwareOffset,
    LegacyLabwareOffsetCreate,
    LabwareOffsetCreate,
    LabwareOffsetVector,
    LegacyLabwareOffsetLocation,
    LabwareOffsetLocationSequence,
    OnLabwareOffsetLocationSequenceComponent,
    OnModuleOffsetLocationSequenceComponent,
    OnAddressableAreaOffsetLocationSequenceComponent,
    LabwareOffsetLocationSequenceComponents,
    LabwareMovementStrategy,
    AddressableOffsetVector,
    DeckPoint,
    DeckType,
    DeckSlotLocation,
    InStackerHopperLocation,
    ModuleLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
    Dimensions,
    EngineStatus,
    LabwareLocation,
    LoadableLabwareLocation,
    NonStackedLocation,
    LoadedLabware,
    LoadedModule,
    LoadedPipette,
    MotorAxis,
    WellLocation,
    LiquidHandlingWellLocation,
    PickUpTipWellLocation,
    DropTipWellLocation,
    WellOrigin,
    DropTipWellOrigin,
    PickUpTipWellOrigin,
    WellOffset,
    ModuleModel,
    ModuleDefinition,
    Liquid,
    LiquidClassRecord,
    LiquidClassRecordWithId,
    AllNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
    RowNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    QuadrantNozzleLayoutConfiguration,
)


__all__ = [
    # main factory and interface exports
    "create_protocol_engine",
    "create_protocol_engine_in_thread",
    "ProtocolEngine",
    "StateSummary",
    "Config",
    # error types
    "ProtocolEngineError",
    "ErrorOccurrence",
    # top level command unions and values
    "Command",
    "CommandParams",
    "CommandCreate",
    "CommandStatus",
    "CommandType",
    "CommandIntent",
    "CommandNote",
    # state interfaces and models
    "State",
    "StateView",
    "CommandSlice",
    "CommandErrorSlice",
    "CommandPointer",
    # public value interfaces and models
    "LabwareOffset",
    "LabwareOffsetCreate",
    "LegacyLabwareOffsetCreate",
    "LabwareOffsetLocationSequence",
    "LabwareOffsetVector",
    "OnLabwareOffsetLocationSequenceComponent",
    "OnModuleOffsetLocationSequenceComponent",
    "OnAddressableAreaOffsetLocationSequenceComponent",
    "LabwareOffsetLocationSequenceComponents",
    "LegacyLabwareOffsetCreate",
    "LegacyLabwareOffsetLocation",
    "LabwareMovementStrategy",
    "AddressableOffsetVector",
    "DeckSlotLocation",
    "DeckPoint",
    "DeckType",
    "ModuleLocation",
    "OnLabwareLocation",
    "AddressableAreaLocation",
    "InStackerHopperLocation",
    "OFF_DECK_LOCATION",
    "SYSTEM_LOCATION",
    "Dimensions",
    "EngineStatus",
    "LabwareLocation",
    "LoadableLabwareLocation",
    "NonStackedLocation",
    "LoadedLabware",
    "LoadedModule",
    "LoadedPipette",
    "MotorAxis",
    "WellLocation",
    "LiquidHandlingWellLocation",
    "PickUpTipWellLocation",
    "DropTipWellLocation",
    "WellOrigin",
    "DropTipWellOrigin",
    "PickUpTipWellOrigin",
    "WellOffset",
    "ModuleModel",
    "ModuleDefinition",
    "Liquid",
    "LiquidClassRecord",
    "LiquidClassRecordWithId",
    "AllNozzleLayoutConfiguration",
    "SingleNozzleLayoutConfiguration",
    "RowNozzleLayoutConfiguration",
    "ColumnNozzleLayoutConfiguration",
    "QuadrantNozzleLayoutConfiguration",
    # plugins
    "AbstractPlugin",
]



"""Protocol engine types dealing with automatic tip selection."""
from enum import Enum
from typing import Optional

from pydantic import (
    BaseModel,
    Field,
)


class NextTipInfo(BaseModel):
    """Next available tip labware and well name data."""

    labwareId: str = Field(
        ...,
        description="The labware ID of the tip rack where the next available tip(s) are located.",
    )
    tipStartingWell: str = Field(
        ..., description="The (starting) well name of the next available tip(s)."
    )


class NoTipReason(Enum):
    """The cause of no tip being available for a pipette and tip rack(s)."""

    NO_AVAILABLE_TIPS = "noAvailableTips"
    STARTING_TIP_WITH_PARTIAL = "startingTipWithPartial"
    INCOMPATIBLE_CONFIGURATION = "incompatibleConfiguration"


class NoTipAvailable(BaseModel):
    """No available next tip data."""

    noTipReason: NoTipReason = Field(
        ..., description="The reason why no next available tip could be provided."
    )
    message: Optional[str] = Field(
        None, description="Optional message explaining why a tip wasn't available."
    )



from typing import Optional
from typing_extensions import Protocol

from opentrons.types import Point
from opentrons.hardware_control.types import CriticalPoint, TipScrapeType
from .types import MountArgType, CalibrationType, ConfigType

from .instrument_configurer import InstrumentConfigurer
from .motion_controller import MotionController
from .configurable import Configurable
from .calibratable import Calibratable


class LiquidHandler(
    InstrumentConfigurer[MountArgType],
    MotionController[MountArgType],
    Configurable[ConfigType],
    Calibratable[CalibrationType],
    Protocol[CalibrationType, MountArgType, ConfigType],
):
    def critical_point_for(
        self,
        mount: MountArgType,
        cp_override: Optional[CriticalPoint] = None,
    ) -> Point:
        """
        Determine the current critical point for the specified mount.

        :param mount: A robot mount that the instrument is on.
        :param cp_override: The critical point override to use.

        If no critical point override is specified, the robot defaults to nozzle location `A1` or the mount critical point.
        :return: Point.
        """
        ...

    async def update_nozzle_configuration_for_mount(
        self,
        mount: MountArgType,
        back_left_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
        starting_nozzle: Optional[str] = None,
    ) -> None:
        """
        The expectation of this function is that the back_left_nozzle/front_right_nozzle are the two corners
        of a rectangle of nozzles. A call to this function that does not follow that schema will result
        in an error.

        :param mount: A robot mount that the instrument is on.
        :param back_left_nozzle: A string representing a nozzle name of the form <LETTER><NUMBER> such as 'A1'.
        :param front_right_nozzle: A string representing a nozzle name of the form <LETTER><NUMBER> such as 'A1'.
        :param starting_nozzle: A string representing the starting nozzle which will be used as the critical point
        of the pipette nozzle configuration. By default, the back left nozzle will be the starting nozzle if
        none is provided.

        If none of the nozzle parameters are provided, the nozzle configuration will be reset to default.
        :return: None.
        """
        ...

    async def configure_for_volume(self, mount: MountArgType, volume: float) -> None:
        """
        Configure a pipette to handle the specified volume.

        Some pipettes need to switch modes to handle different volumes of liquid. Calling
        this ensures they do so. If the pipette does not need to switch modes, or the pipette
        is already in the specified mode, nothing will happen.

        If the pipette does need to switch modes, it will likely be unready for aspiration and
        you will need to call :py:meth:`prepare_for_aspirate` afterwards.
        """
        ...

    async def prepare_for_aspirate(
        self, mount: MountArgType, rate: float = 1.0
    ) -> None:
        """
        Prepare the pipette for aspiration.

        This must happen after every :py:meth:`blow_out` and should probably be
        called before every :py:meth:`aspirate`, while the pipette tip is not
        immersed in a well. It ensures that the plunger is at the 0-volume
        position of the pipette if necessary (if not necessary, it does
        nothing).

        If :py:meth:`aspirate` is called immediately after :py:meth:`blow_out`,
        the plunger is left at the ``blow_out`` position, below the ``bottom``
        position, and moving the plunger up during :py:meth:`aspirate` is
        expected to aspirate liquid - :py:meth:`aspirate` is called once the
        pipette tip is already in the well. This will cause a subtle over
        aspiration. To make the problem more obvious, :py:meth:`aspirate` will
        raise an exception if this method has not previously been called.
        """
        ...

    async def aspirate(
        self,
        mount: MountArgType,
        volume: Optional[float] = None,
        rate: float = 1.0,
        correction_volume: float = 0.0,
    ) -> None:
        """
        Aspirate a volume of liquid (in microliters/uL) using this pipette
        from the *current location*. If no volume is passed, `aspirate` will
        default to max available volume (after taking into account the volume
        already present in the tip).

        The function :py:meth:`prepare_for_aspirate` must be called prior to
        calling this function, while the tip is above the well. This ensures
        that the pipette tip is in the proper position at the bottom of the
        pipette to begin aspiration, and prevents subtle over-aspiration if
        an aspirate is done immediately after :py:meth:`blow_out`. If
        :py:meth:`prepare_for_aspirate` has not been called since the last
        call to :py:meth:`aspirate`, an exception will be raised.

        mount : Mount.LEFT or Mount.RIGHT
        volume : [float] The number of microliters to aspirate
        rate : [float] Set plunger speed for this aspirate, where
            speed = rate * aspirate_speed
        correction_volume : Correction volume in uL for the specified aspirate volume
        """
        ...

    async def aspirate_while_tracking(
        self,
        mount: MountArgType,
        z_distance: float,
        volume: float,
        flow_rate: float = 1.0,
    ) -> None:
        """
        Aspirate a volume of liquid (in microliters/uL) while moving the z axis synchronously.

        :param mount: A robot mount that the instrument is on.
        :param z_distance: The distance the z axis will move during apsiration.
        :param volume: The volume of liquid to be aspirated.
        :param flow_rate: The flow rate to aspirate with.
        """
        ...

    async def dispense(
        self,
        mount: MountArgType,
        volume: Optional[float] = None,
        rate: float = 1.0,
        push_out: Optional[float] = None,
        correction_volume: float = 0.0,
        is_full_dispense: bool = False,
    ) -> None:
        """
        Dispense a volume of liquid in microliters(uL) using this pipette
        at the current location. If no volume is specified, `dispense` will
        dispense all volume currently present in pipette

        mount : Mount.LEFT or Mount.RIGHT
        volume : [float] The number of microliters to dispense
        rate : [float] Set plunger speed for this dispense, where
            speed = rate * dispense_speed
        correction_volume : Correction volume in uL for the specified dispense volume
        """
        ...

    async def dispense_while_tracking(
        self,
        mount: MountArgType,
        z_distance: float,
        volume: float,
        push_out: Optional[float],
        flow_rate: float = 1.0,
        is_full_dispense: bool = False,
    ) -> None:
        """
        Dispense a volume of liquid (in microliters/uL) while moving the z axis synchronously.

        :param mount: A robot mount that the instrument is on.
        :param z_distance: The distance the z axis will move during dispensing.
        :param volume: The volume of liquid to be dispensed.
        :param flow_rate: The flow rate to dispense with.
        """
        ...

    async def blow_out(
        self, mount: MountArgType, volume: Optional[float] = None
    ) -> None:
        """
        Force any remaining liquid to dispense. The liquid will be dispensed at
        the current location of pipette
        """
        ...

    async def tip_pickup_moves(
        self,
        mount: MountArgType,
        presses: Optional[int] = None,
        increment: Optional[float] = None,
    ) -> None:
        ...

    async def pick_up_tip(
        self,
        mount: MountArgType,
        tip_length: float,
        presses: Optional[int] = None,
        increment: Optional[float] = None,
        prep_after: bool = True,
    ) -> None:
        """
        Pick up tip from current location.

        This is achieved by attempting to move the instrument down by its
        `pick_up_distance`, in a series of presses. This distance is larger
        than the space available in the tip, so the stepper motor will
        eventually skip steps, which is resolved by homing afterwards. The
        pick up operation is done at a current specified in the pipette config,
        which is experimentally determined to skip steps at a level of force
        sufficient to provide a good seal between the pipette nozzle and tip
        while also avoiding attaching the tip so firmly that it can't be
        dropped later.

        If ``presses`` or ``increment`` is not specified (or is ``None``),
        their value is taken from the pipette configuration.
        """
        ...

    async def tip_drop_moves(
        self,
        mount: MountArgType,
        home_after: bool = True,
        ignore_plunger: bool = False,
        scrape_type: TipScrapeType = TipScrapeType.NONE,
    ) -> None:
        ...

    async def drop_tip(
        self,
        mount: MountArgType,
        home_after: bool = True,
    ) -> None:
        """
        Drop tip at the current location

        :param Mount mount: The mount to drop a tip from
        :param bool home_after: Home the plunger motor after dropping tip. This
                                is used in case the plunger motor skipped while
                                dropping the tip, and is also used to recover
                                the ejector shroud after a drop.
        """
        ...

    async def liquid_probe(
        self,
        mount: MountArgType,
        max_z_dist: float,
    ) -> float:
        """Search for and return liquid level height using this pipette
        at the current location.

        mount : Mount.LEFT or Mount.RIGHT
        max_z_dist : maximum depth to probe for liquid
        """
        ...

    async def increase_evo_disp_count(self, mount: MountArgType) -> None:
        """Tell a pipette to increase it's evo-tip-dispense-count in eeprom."""
        ...



"""Shared result types for robot API commands."""
from pydantic import BaseModel, Field

from typing import Dict
from opentrons.protocol_engine.types import MotorAxis


MotorAxisMapType = Dict[MotorAxis, float]
default_position = {ax: 0.0 for ax in MotorAxis}


class DestinationRobotPositionResult(BaseModel):
    """The result dictionary of `MotorAxis` type."""

    position: MotorAxisMapType = Field(
        default=default_position,
        description="The position of all axes on the robot. If no mount was provided, the last moved mount is used to determine the location.",
    )



"""Abstract interfaces for engine plugins."""
from abc import ABC, abstractmethod

from .actions import Action


class ActionHandler(ABC):
    """An abstract interface for an object that reacts to actions."""

    @abstractmethod
    def handle_action(self, action: Action) -> None:
        """React to a state-change action."""
        ...



import logging
from typing import Optional, Any

from opentrons_shared_data.labware.types import (
    LabwareDefinition2 as LabwareDefinition2Dict,
)

from opentrons import types
from opentrons.calibration_storage.helpers import uri_from_definition
from opentrons.calibration_storage.types import TipLengthCalNotFound
from opentrons.hardware_control.dev_types import PipetteDict

# TODO (lc 09-26-2022) We should conditionally import ot2 or ot3 calibration
from opentrons.hardware_control.instruments.ot2 import (
    instrument_calibration as instr_cal,
)
from opentrons.protocol_api.labware import Labware
from opentrons.protocols.api_support.types import APIVersion
from opentrons_shared_data.protocol.types import (
    LiquidHandlingCommand,
    BlowoutLocation,
)


def validate_blowout_location(
    api_version: APIVersion,
    liquid_handling_command: LiquidHandlingCommand,
    blowout_location: Optional[Any],
) -> None:
    """Validate the blowout location."""
    if blowout_location and api_version < APIVersion(2, 8):
        raise ValueError(
            "Cannot specify blowout location when using api"
            + " version below 2.8, current version is {api_version}".format(
                api_version=api_version
            )
        )

    elif liquid_handling_command == "consolidate" and blowout_location == "source well":
        raise ValueError("blowout location for consolidate cannot be source well")
    elif (
        liquid_handling_command == "distribute"
        and blowout_location == "destination well"
    ):
        raise ValueError("blowout location for distribute cannot be destination well")
    elif (
        liquid_handling_command == "transfer"
        and blowout_location
        and blowout_location not in [location.value for location in BlowoutLocation]
    ):
        raise ValueError(
            "blowout location should be either 'source well', "
            + " 'destination well', or 'trash'"
            + f" but it is {blowout_location}"
        )


def tip_length_for(
    pipette: PipetteDict, tip_rack_definition: LabwareDefinition2Dict
) -> float:
    """Get the tip length, including overlap, for a tip from this rack"""
    try:
        return instr_cal.load_tip_length_for_pipette(
            pipette["pipette_id"], tip_rack_definition
        ).tip_length
    except TipLengthCalNotFound:
        tip_overlap = pipette["tip_overlap"].get(
            uri_from_definition(tip_rack_definition),
            pipette["tip_overlap"]["default"],
        )
        tip_length = tip_rack_definition["parameters"]["tipLength"]
        return tip_length - tip_overlap


VALID_PIP_TIPRACK_VOL = {
    "FLEX": {"p50": [50], "p200": [50, 200], "p1000": [50, 200, 1000]},
    "OT2": {
        "p10": [10, 20],
        "p20": [10, 20],
        "p50": [50, 200, 300],
        "p300": [200, 300],
        "p1000": [1000],
    },
}


def validate_tiprack(
    instrument_name: str, tip_rack: Labware, log: logging.Logger
) -> None:
    """Validate a tiprack logging a warning message."""
    if not tip_rack.is_tiprack:
        raise ValueError(f"Labware {tip_rack.load_name} is not a tip rack.")

    # TODO AA 2020-06-24 - we should instead add the acceptable Opentrons
    #  tipracks to the pipette as a refactor
    if tip_rack.uri.startswith("opentrons/"):
        tiprack_vol = tip_rack.wells()[0].max_volume
        instr_metadata = instrument_name.split("_")
        gen_lookup = (
            "FLEX" if ("flex" in instr_metadata or "96" in instr_metadata) else "OT2"
        )

        # TODO (spp, 2025-01-30): do what AA's note above says or at least,
        #  fetch the 'pip_type' below from the 'model' field in pipette definitions
        #  so that we don't have to figure it out from pipette names
        if instrument_name.split("_")[0] == "flex":
            # Flex's API load names have the format 'flex_1channel_1000'
            # From API v2.23 on, this is the name returned by InstrumentContext.name
            pip_type = "p" + instrument_name.split("_")[2]
        else:
            # Until API v2.23, InstrumentContext.name returned the engine-specific names
            # of Flex pipettes. These names, as well as OT2 pipette names,
            # have the format- 'p1000_single_gen2' or 'p1000_single_flex'
            pip_type = instrument_name.split("_")[0]
        valid_vols = VALID_PIP_TIPRACK_VOL[gen_lookup][pip_type]
        if tiprack_vol not in valid_vols:
            log.warning(
                f"The pipette {instrument_name} and its tip rack {tip_rack.load_name}"
                " appear to be mismatched. Please check your protocol."
            )


def validate_takes_liquid(
    location: types.Location, reject_module: bool, reject_adapter: bool
) -> None:
    """Validate that a location is a valid liquid handling target.

    Args:
        location: target location.
        reject_module: if set to True, disallow liquid handling directly to module
        reject_adapter: if set to True, check that labware is not an adapter

    Raises:
        ValueError: the given location is not a valid liquid handling target.
    """
    labware = None

    if location.labware.is_labware:
        labware = location.labware.as_labware()

    if location.labware.is_well:
        labware = location.labware.as_well().parent

    if location.labware.is_module and reject_module:
        raise ValueError("Cannot aspirate/dispense directly to a module")

    if labware is not None and labware.is_tiprack:
        raise ValueError("Cannot aspirate/dispense to a tip rack")

    if reject_adapter and (labware is not None and labware.is_adapter):
        raise ValueError("Cannot aspirate/dispense to an adapter")



import os

from pathlib import Path
import logging
import re
from typing import Any, List, Tuple

from opentrons.drivers.serial_communication import get_ports_by_name
from opentrons.hardware_control import (
    API as HardwareAPI,
    ThreadManager,
    ThreadManagedHardware,
    types as hw_types,
)

from opentrons.config import (
    feature_flags as ff,
    name,
    robot_configs,
    IS_ROBOT,
    ROBOT_FIRMWARE_DIR,
)
from opentrons.util import logging_config
from opentrons.protocols.types import ApiDeprecationError
from opentrons.protocols.api_support.types import APIVersion

from ._version import version

HERE = os.path.abspath(os.path.dirname(__file__))
__version__ = version


LEGACY_MODULES = ["robot", "reset", "instruments", "containers", "labware", "modules"]


__all__ = ["version", "__version__", "HERE", "config"]


def __getattr__(attrname: str) -> None:
    """
    Prevent import of legacy modules from global to officially
    deprecate Python API Version 1.0.
    """
    if attrname in LEGACY_MODULES:
        raise ApiDeprecationError(APIVersion(1, 0))
    raise AttributeError(attrname)


def __dir__() -> List[str]:
    return sorted(__all__ + LEGACY_MODULES)


log = logging.getLogger(__name__)


SMOOTHIE_HEX_RE = re.compile("smoothie-(.*).hex")


def _find_smoothie_file() -> Tuple[Path, str]:
    resources: List[Path] = []

    # Search for smoothie files in /usr/lib/firmware first then fall back to
    # value packed in wheel
    if IS_ROBOT:
        resources.extend(ROBOT_FIRMWARE_DIR.iterdir())  # type: ignore

    resources_path = Path(HERE) / "resources"
    resources.extend(resources_path.iterdir())

    for path in resources:
        matches = SMOOTHIE_HEX_RE.search(path.name)
        if matches:
            branch_plus_ref = matches.group(1)
            return path, branch_plus_ref
    raise OSError(f"Could not find smoothie firmware file in {resources_path}")


def _get_motor_control_serial_port() -> Any:
    port = os.environ.get("OT_SMOOTHIE_EMULATOR_URI")

    if port is None:
        smoothie_id = os.environ.get("OT_SMOOTHIE_ID", "AMA")
        # TODO(mc, 2021-08-01): raise a more informative exception than
        # IndexError if a valid serial port is not found
        port = get_ports_by_name(device_name=smoothie_id)[0]

    log.info(f"Connecting to motor controller at port {port}")
    return port


def should_use_ot3() -> bool:
    """Return true if ot3 hardware controller should be used."""
    if ff.enable_ot3_hardware_controller():
        try:
            from opentrons_hardware.drivers.can_bus import CanDriver  # noqa: F401

            return True
        except ModuleNotFoundError:
            log.exception("Cannot use OT3 Hardware controller.")
    return False


async def _create_thread_manager() -> ThreadManagedHardware:
    """Build the hardware controller wrapped in a ThreadManager.

    .. deprecated:: 4.6
        ThreadManager is on its way out.
    """
    if os.environ.get("ENABLE_VIRTUAL_SMOOTHIE"):
        log.info("Initialized robot using virtual Smoothie")
        thread_manager: ThreadManagedHardware = ThreadManager(
            HardwareAPI.build_hardware_simulator
        )
    elif should_use_ot3():
        from opentrons.hardware_control.ot3api import OT3API

        thread_manager = ThreadManager(
            ThreadManager.nonblocking_builder(OT3API.build_hardware_controller),
            use_usb_bus=ff.rear_panel_integration(),
            status_bar_enabled=ff.status_bar_enabled(),
            feature_flags=hw_types.HardwareFeatureFlags.build_from_ff(),
        )
    else:
        thread_manager = ThreadManager(
            ThreadManager.nonblocking_builder(HardwareAPI.build_hardware_controller),
            port=_get_motor_control_serial_port(),
            firmware=_find_smoothie_file(),
            feature_flags=hw_types.HardwareFeatureFlags.build_from_ff(),
        )

    try:
        await thread_manager.managed_thread_ready_async()
    except RuntimeError:
        log.exception("Could not build hardware controller, forcing virtual")
        thread_manager = ThreadManager(HardwareAPI.build_hardware_simulator)

    return thread_manager


async def initialize() -> ThreadManagedHardware:
    """
    Initialize the Opentrons hardware returning a hardware instance.
    """
    robot_conf = robot_configs.load()
    logging_config.log_init(robot_conf.log_level)

    log.info(f"API server version: {version}")
    log.info(f"Robot Name: {name()}")

    return await _create_thread_manager()



"""The interface that implements InstrumentContext."""

from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Any, Generic, Optional, TypeVar, Union, List, Tuple, Literal

from opentrons import types
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support.util import FlowRates
from opentrons.protocols.advanced_control.transfers.common import TransferTipPolicyV2
from opentrons.protocol_api._nozzle_layout import NozzleLayout
from opentrons.protocol_api._liquid import LiquidClass
from opentrons.protocol_engine.types import LiquidTrackingType

from ..disposal_locations import TrashBin, WasteChute
from .well import WellCoreType
from .labware import LabwareCoreType


class AbstractInstrument(ABC, Generic[WellCoreType, LabwareCoreType]):
    @abstractmethod
    def get_default_speed(self) -> float:
        ...

    @abstractmethod
    def set_default_speed(self, speed: float) -> None:
        ...

    @abstractmethod
    def air_gap_in_place(
        self, volume: float, flow_rate: float, correction_volume: Optional[float] = None
    ) -> None:
        """Aspirate a given volume of air from the current location of the pipette.
        Args:
            volume: The volume of air to aspirate, in microliters.
            flow_rate: The flow rate of air into the pipette, in microliters.
            correction_volume: The correction volume in uL.
        """

    @abstractmethod
    def aspirate(
        self,
        location: types.Location,
        well_core: Optional[WellCoreType],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
        meniscus_tracking: Optional[types.MeniscusTrackingTarget] = None,
        correction_volume: Optional[float] = None,
    ) -> None:
        """Aspirate a given volume of liquid from the specified location.
        Args:
            volume: The volume of liquid to aspirate, in microliters.
            location: The exact location to aspirate from.
            well_core: The well to aspirate from, if applicable.
            rate: The rate for how quickly to aspirate.
            flow_rate: The flow rate in µL/s to aspirate at.
            in_place: Whether this is in-place.
            meniscus_tracking: Optional data about where to aspirate from.
            correction_volume: The correction volume in uL
        """
        ...

    @abstractmethod
    def dispense(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[WellCoreType],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
        push_out: Optional[float],
        meniscus_tracking: Optional[types.MeniscusTrackingTarget] = None,
        correction_volume: Optional[float] = None,
    ) -> None:
        """Dispense a given volume of liquid into the specified location.
        Args:
            volume: The volume of liquid to dispense, in microliters.
            location: The exact location to dispense to.
            well_core: The well to dispense to, if applicable.
            rate: The rate for how quickly to dispense.
            flow_rate: The flow rate in µL/s to dispense at.
            in_place: Whether this is in-place.
            push_out: The amount to push the plunger below bottom position.
            correction_volume: The correction volume in uL
            meniscus_tracking: Optional data about where to dispense from.
        """
        ...

    @abstractmethod
    def blow_out(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[WellCoreType],
        in_place: bool,
    ) -> None:
        """Blow liquid out of the tip.

        Args:
            location: The location to blow out into.
            well_core: The well to blow out into.
            in_place: Whether this is in-place.
        """
        ...

    @abstractmethod
    def touch_tip(
        self,
        location: types.Location,
        well_core: WellCoreType,
        radius: float,
        z_offset: float,
        speed: float,
        mm_from_edge: Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def pick_up_tip(
        self,
        location: types.Location,
        well_core: WellCoreType,
        presses: Optional[int],
        increment: Optional[float],
        prep_after: bool = True,
    ) -> None:
        """Move to and pick up a tip from a given well.

        Args:
            location: The location of the well we're picking up from.
            well_core: The well to pick up from.
            presses: Customize the number of presses the pipette does.
            increment: Customize the movement "distance" of the pipette to press harder.
            prep_after: Move plunger to the "ready to aspirate" position after pick up.
        """
        ...

    @abstractmethod
    def drop_tip(
        self,
        location: Optional[types.Location],
        well_core: WellCoreType,
        home_after: Optional[bool],
        alternate_drop_location: Optional[bool] = False,
    ) -> None:
        """Move to and drop a tip into a given well.

        Args:
            location: The location of the well we're dropping to.
                If unspecified, the default drop location of the well will be used.
            well_core: The well we're dropping into
            home_after: Whether to home the pipette after the tip is dropped.
            alternate_drop_location: Whether to randomize the exact location to drop tip
                within the specified well.
        """
        ...

    @abstractmethod
    def drop_tip_in_disposal_location(
        self,
        disposal_location: Union[TrashBin, WasteChute],
        home_after: Optional[bool],
        alternate_tip_drop: bool = False,
    ) -> None:
        """Move to and drop tip into a TrashBin or WasteChute.

        Args:
            disposal_location: The disposal location object we're dropping to.
            home_after: Whether to home the pipette after the tip is dropped.
            alternate_tip_drop: Whether to alternate tip drop location in a trash bin.
        """
        ...

    @abstractmethod
    def home(self) -> None:
        ...

    @abstractmethod
    def home_plunger(self) -> None:
        ...

    @abstractmethod
    def move_to(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[WellCoreType],
        force_direct: bool,
        minimum_z_height: Optional[float],
        speed: Optional[float],
        check_for_movement_conflicts: bool,
    ) -> None:
        ...

    @abstractmethod
    def resin_tip_seal(
        self,
        location: types.Location,
        well_core: WellCoreType,
        in_place: Optional[bool] = False,
    ) -> None:
        ...

    @abstractmethod
    def resin_tip_unseal(
        self,
        location: types.Location | None,
        well_core: WellCoreType,
    ) -> None:
        ...

    @abstractmethod
    def resin_tip_dispense(
        self,
        location: types.Location,
        well_core: WellCoreType,
        volume: Optional[float] = None,
        flow_rate: Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def get_mount(self) -> types.Mount:
        ...

    @abstractmethod
    def get_pipette_name(self) -> str:
        ...

    @abstractmethod
    def get_model(self) -> str:
        ...

    @abstractmethod
    def get_display_name(self) -> str:
        ...

    @abstractmethod
    def get_min_volume(self) -> float:
        ...

    @abstractmethod
    def get_max_volume(self) -> float:
        ...

    @abstractmethod
    def get_working_volume(self) -> float:
        ...

    @abstractmethod
    def get_current_volume(self) -> float:
        ...

    @abstractmethod
    def get_has_clean_tip(self) -> bool:
        ...

    @abstractmethod
    def get_available_volume(self) -> float:
        ...

    @abstractmethod
    def get_minimum_liquid_sense_height(self) -> float:
        ...

    @abstractmethod
    def get_hardware_state(self) -> PipetteDict:
        """Get the current state of the pipette hardware as a dictionary."""
        ...

    @abstractmethod
    def get_channels(self) -> int:
        ...

    @abstractmethod
    def get_active_channels(self) -> int:
        ...

    @abstractmethod
    def get_nozzle_map(self) -> types.NozzleMapInterface:
        ...

    @abstractmethod
    def has_tip(self) -> bool:
        ...

    @abstractmethod
    def get_return_height(self) -> float:
        ...

    @abstractmethod
    def get_flow_rate(self) -> FlowRates:
        ...

    @abstractmethod
    def get_aspirate_flow_rate(self, rate: float = 1.0) -> float:
        ...

    @abstractmethod
    def get_dispense_flow_rate(self, rate: float = 1.0) -> float:
        ...

    @abstractmethod
    def get_blow_out_flow_rate(self, rate: float = 1.0) -> float:
        ...

    @abstractmethod
    def get_liquid_presence_detection(self) -> bool:
        ...

    @abstractmethod
    def _pressure_supported_by_pipette(self) -> bool:
        ...

    @abstractmethod
    def set_liquid_presence_detection(self, enable: bool) -> None:
        ...

    @abstractmethod
    def set_flow_rate(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def configure_for_volume(self, volume: float) -> None:
        """Configure the pipette for a specific volume.

        Args:
            volume: The volume to prepare to handle.
        """
        ...

    @abstractmethod
    def prepare_to_aspirate(self) -> None:
        """Prepare the pipette to aspirate."""
        ...

    @abstractmethod
    def configure_nozzle_layout(
        self,
        style: NozzleLayout,
        primary_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
        back_left_nozzle: Optional[str],
    ) -> None:
        """Configure the pipette to a specific nozzle layout.

        Args:
            style: The type of configuration you wish to build.
            primary_nozzle: The nozzle that will determine a pipette's critical point.
            front_right_nozzle: The front right most nozzle in the requested layout.
            back_left_nozzle: The back left most nozzle in the requested layout.
        """
        ...

    @abstractmethod
    def transfer_with_liquid_class(
        self,
        liquid_class: LiquidClass,
        volume: float,
        source: List[Tuple[types.Location, WellCoreType]],
        dest: Union[List[Tuple[types.Location, WellCoreType]], TrashBin, WasteChute],
        new_tip: TransferTipPolicyV2,
        tip_racks: List[Tuple[types.Location, LabwareCoreType]],
        starting_tip: Optional[WellCoreType],
        trash_location: Union[types.Location, TrashBin, WasteChute],
        return_tip: bool,
        keep_last_tip: bool,
        last_tip_location: Optional[Tuple[types.Location, WellCoreType]],
    ) -> Optional[Tuple[types.Location, WellCoreType]]:
        """Transfer a liquid from source to dest according to liquid class properties."""
        ...

    @abstractmethod
    def distribute_with_liquid_class(
        self,
        liquid_class: LiquidClass,
        volume: float,
        source: Tuple[types.Location, WellCoreType],
        dest: List[Tuple[types.Location, WellCoreType]],
        new_tip: Literal[TransferTipPolicyV2.NEVER, TransferTipPolicyV2.ONCE],
        tip_racks: List[Tuple[types.Location, LabwareCoreType]],
        starting_tip: Optional[WellCoreType],
        trash_location: Union[types.Location, TrashBin, WasteChute],
        return_tip: bool,
        keep_last_tip: bool,
        last_tip_location: Optional[Tuple[types.Location, WellCoreType]],
    ) -> Optional[Tuple[types.Location, WellCoreType]]:
        """
        Distribute a liquid from single source to multiple destinations
        according to liquid class properties.
        """
        ...

    @abstractmethod
    def consolidate_with_liquid_class(
        self,
        liquid_class: LiquidClass,
        volume: float,
        source: List[Tuple[types.Location, WellCoreType]],
        dest: Union[Tuple[types.Location, WellCoreType], TrashBin, WasteChute],
        new_tip: Literal[TransferTipPolicyV2.NEVER, TransferTipPolicyV2.ONCE],
        tip_racks: List[Tuple[types.Location, LabwareCoreType]],
        starting_tip: Optional[WellCoreType],
        trash_location: Union[types.Location, TrashBin, WasteChute],
        return_tip: bool,
        keep_last_tip: bool,
        last_tip_location: Optional[Tuple[types.Location, WellCoreType]],
    ) -> Optional[Tuple[types.Location, WellCoreType]]:
        """
        Consolidate liquid from multiple sources to a single destination
        using the specified liquid class properties.
        """
        ...

    @abstractmethod
    def is_tip_tracking_available(self) -> bool:
        """Return whether auto tip tracking is available for the pipette's current nozzle configuration."""

    @abstractmethod
    def retract(self) -> None:
        """Retract this instrument to the top of the gantry."""
        ...

    @abstractmethod
    def detect_liquid_presence(
        self, well_core: WellCoreType, loc: types.Location
    ) -> bool:
        """Do a liquid probe to detect whether there is liquid in the well."""

    @abstractmethod
    def liquid_probe_with_recovery(
        self, well_core: WellCoreType, loc: types.Location
    ) -> None:
        """Do a liquid probe to detect the presence of liquid in the well."""
        ...

    @abstractmethod
    def liquid_probe_without_recovery(
        self, well_core: WellCoreType, loc: types.Location
    ) -> LiquidTrackingType:
        """Do a liquid probe to find the level of the liquid in the well."""
        ...

    @abstractmethod
    def nozzle_configuration_valid_for_lld(self) -> bool:
        """Check if the nozzle configuration currently supports LLD."""


InstrumentCoreType = TypeVar("InstrumentCoreType", bound=AbstractInstrument[Any, Any])
